"""Defines the Op protocol and related functions."""

from __future__ import annotations

import atexit
import inspect
import logging
import random
import sys
import traceback
import weakref
from collections.abc import AsyncIterator, Coroutine, Generator, Iterator, Mapping
from dataclasses import dataclass
from functools import partial, wraps
from types import MethodType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Optional,
    Protocol,
    TypedDict,
    TypeVar,
    cast,
    overload,
    runtime_checkable,
)

from weave.trace import box, settings
from weave.trace.constants import TRACE_CALL_EMOJI
from weave.trace.context import call_context
from weave.trace.context import weave_client_context as weave_client_context
from weave.trace.context.call_context import (
    call_attributes,
    get_tracing_enabled,
    tracing_disabled,
)
from weave.trace.context.tests_context import get_raise_on_captured_errors
from weave.trace.refs import ObjectRef
from weave.trace.util import log_once

if TYPE_CHECKING:
    from weave.trace.weave_client import Call, CallsIter

try:
    from openai._types import NOT_GIVEN as OPENAI_NOT_GIVEN
except ImportError:
    OPENAI_NOT_GIVEN = None

try:
    from cohere.base_client import COHERE_NOT_GIVEN
except ImportError:
    COHERE_NOT_GIVEN = None

try:
    from anthropic._types import NOT_GIVEN as ANTHROPIC_NOT_GIVEN
except ImportError:
    ANTHROPIC_NOT_GIVEN = None

try:
    from cerebras.cloud.sdk._types import NOT_GIVEN as CEREBRAS_NOT_GIVEN
except ImportError:
    CEREBRAS_NOT_GIVEN = None


S = TypeVar("S")
V = TypeVar("V")


if sys.version_info < (3, 10):

    def aiter(obj: AsyncIterator[V]) -> AsyncIterator[V]:
        return obj.__aiter__()

    async def anext(obj: AsyncIterator[V], default: Optional[V] = None) -> V:  # noqa: UP007,UP045
        try:
            return await obj.__anext__()
        except StopAsyncIteration:
            if default is not None:
                return default
            else:
                raise


logger = logging.getLogger(__name__)


CALL_CREATE_MSG = "Error creating call:\n{}"
ASYNC_CALL_CREATE_MSG = "Error creating async call:\n{}"
ON_OUTPUT_MSG = "Error capturing call output:\n{}"
UNINITIALIZED_MSG = "Warning: Traces will not be logged. Call weave.init to log your traces to a project.\n"


class DisplayNameFuncError(ValueError): ...


def print_call_link(call: Call) -> None:
    if settings.should_print_call_link():
        print(f"{TRACE_CALL_EMOJI} {call.ui_url}")


@dataclass
class ProcessedInputs:
    # What the user passed to the function
    original_args: tuple
    original_kwargs: dict[str, Any]

    # What should get passed to the interior function
    args: tuple
    kwargs: dict[str, Any]

    # What should get sent to the Weave server
    inputs: dict[str, Any]


OnInputHandlerType = Callable[["Op", tuple, dict], Optional[ProcessedInputs]]
FinishCallbackType = Callable[[Any, Optional[BaseException]], None]
OnOutputHandlerType = Callable[[Any, FinishCallbackType, dict], Any]
# Call, original function output, exception if occurred
OnFinishHandlerType = Callable[["Call", Any, Optional[BaseException]], None]


def _value_is_sentinel(param: Any) -> bool:
    return param.default in (
        None,
        Ellipsis,
        OPENAI_NOT_GIVEN,
        COHERE_NOT_GIVEN,
        ANTHROPIC_NOT_GIVEN,
        CEREBRAS_NOT_GIVEN,
    )


def _apply_fn_defaults_to_inputs(
    fn: Callable, inputs: Mapping[str, Any]
) -> dict[str, Any]:
    inputs = {**inputs}
    sig = inspect.signature(fn)
    for name, param in sig.parameters.items():
        if name in inputs:
            continue
        if param.default != inspect.Parameter.empty and not _value_is_sentinel(param):
            inputs[name] = param.default
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            inputs[name] = ()
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            inputs[name] = {}
    return inputs


class WeaveKwargs(TypedDict):
    display_name: str | None


@runtime_checkable
class Op(Protocol):
    """
    The interface for Op-ified functions and methods.

    Op was previously a class, and has been converted to a Protocol to allow
    functions to pass for Op.  This is needed because many popular packages are
    using the `inspect` module for control flow, and Op instances don't always
    pass those checks.  In particular, `inspect.iscoroutinefunction` always
    fails for classes, even ones that implement async methods or protocols.

    Some of the attributes are carry-overs from when Op was a class.  We should
    consider removing the unnecessary ones where possible.
    - resolve_fn (I think you can just use the func itself?)
    - _set_on_output_handler (does this have to be on the op?)
    - _on_output_handler (does this have to be on the op?)
    """

    name: str
    call_display_name: str | Callable[[Call], str]
    ref: ObjectRef | None
    resolve_fn: Callable

    postprocess_inputs: Callable[[dict[str, Any]], dict[str, Any]] | None
    postprocess_output: Callable[..., Any] | None

    call: Callable[..., Any]
    calls: Callable[..., CallsIter]

    _set_on_input_handler: Callable[[OnInputHandlerType], None]
    _on_input_handler: OnInputHandlerType | None

    # not sure if this is the best place for this, but kept for compat
    _set_on_output_handler: Callable[[OnOutputHandlerType], None]
    _on_output_handler: OnOutputHandlerType | None

    _set_on_finish_handler: Callable[[OnFinishHandlerType], None]
    _on_finish_handler: OnFinishHandlerType | None

    __call__: Callable[..., Any]
    __self__: Any

    # `_tracing_enabled` is a runtime-only flag that can be used to disable
    # call tracing for an op. It is not persisted as a property of the op, but is
    # respected by the current execution context. It is an underscore property
    # because it is not intended to be used by users directly, but rather assists
    # with internal Weave behavior. If we find a need to expose this to users, we
    # should consider a more user-friendly API (perhaps a setter/getter) & whether
    # it disables child ops as well.
    _tracing_enabled: bool

    tracing_sample_rate: float


def _set_on_input_handler(func: Op, on_input: OnInputHandlerType) -> None:
    if func._on_input_handler is not None:
        raise ValueError("Cannot set on_input_handler multiple times")
    func._on_input_handler = on_input


def _set_on_output_handler(func: Op, on_output: OnOutputHandlerType) -> None:
    if func._on_output_handler is not None:
        raise ValueError("Cannot set on_output_handler multiple times")
    func._on_output_handler = on_output


def _set_on_finish_handler(func: Op, on_finish: OnFinishHandlerType) -> None:
    if func._on_finish_handler is not None:
        raise ValueError("Cannot set on_finish_handler multiple times")
    func._on_finish_handler = on_finish


def _is_unbound_method(func: Callable) -> bool:
    """Check if a function is a function defined on a class (an "unbound" method)

    In python3, the "unbound" method is just a function, but that distinction is
    not enough for our decorator because it needs to operate on both regular funcs
    and unbound methods at the same time.

    This check clarifies that distinction between function vs. unbound method.
    """
    sig = inspect.signature(func)
    params = list(sig.parameters.values())
    is_method = params and params[0].name in {"self", "cls"}

    return bool(is_method)


class OpCallError(Exception): ...


def _default_on_input_handler(func: Op, args: tuple, kwargs: dict) -> ProcessedInputs:
    try:
        sig = inspect.signature(func)
        inputs = sig.bind(*args, **kwargs).arguments
    except TypeError as e:
        raise OpCallError(f"Error calling {func.name}: {e}")

    inputs_with_defaults = _apply_fn_defaults_to_inputs(func, inputs)
    return ProcessedInputs(
        original_args=args,
        original_kwargs=kwargs,
        args=args,
        kwargs=kwargs,
        inputs=inputs_with_defaults,
    )


def _create_call(
    func: Op, *args: Any, __weave: WeaveKwargs | None = None, **kwargs: Any
) -> Call:
    client = weave_client_context.require_weave_client()

    pargs = None
    if func._on_input_handler is not None:
        pargs = func._on_input_handler(func, args, kwargs)
    if not pargs:
        pargs = _default_on_input_handler(func, args, kwargs)
    inputs_with_defaults = pargs.inputs

    # This should probably be configurable, but for now we redact the api_key
    if "api_key" in inputs_with_defaults:
        inputs_with_defaults["api_key"] = "REDACTED"

    call_time_display_name = __weave.get("display_name") if __weave else None

    # If/When we do memoization, this would be a good spot

    parent_call = call_context.get_current_call()
    attributes = call_attributes.get()
    from weave.trace.serialization.serialize import dictify

    attributes = dictify(attributes)

    return client.create_call(
        func,
        inputs_with_defaults,
        parent_call,
        # Very important for `call_time_display_name` to take precedence over `func.call_display_name`
        display_name=call_time_display_name or func.call_display_name,
        attributes=attributes,
    )


def is_tracing_setting_disabled() -> bool:
    if settings.should_disable_weave():
        return True
    if weave_client_context.get_weave_client() is None:
        log_once(logger.warn, UNINITIALIZED_MSG)
        return True
    if not get_tracing_enabled():
        return True
    return False


def should_skip_tracing_for_op(op: Op) -> bool:
    return not op._tracing_enabled


def _should_sample_traces(op: Op) -> bool:
    if call_context.get_current_call():
        return False  # Don't sample traces for child calls

    if random.random() > op.tracing_sample_rate:
        return True  # Sample traces for this call

    return False


def placeholder_call() -> Call:
    # Import here to avoid circular dependency
    from weave.trace.weave_client import NoOpCall

    return NoOpCall()


def is_placeholder_call(call: Call) -> bool:
    from weave.trace.weave_client import NoOpCall

    return isinstance(call, NoOpCall)


def _call_sync_func(
    op: Op,
    *args: Any,
    __weave: WeaveKwargs | None = None,
    __should_raise: bool = False,
    **kwargs: Any,
) -> tuple[Any, Call]:
    func = op.resolve_fn
    call = placeholder_call()

    # Handle all of the possible cases where we would skip tracing.
    if is_tracing_setting_disabled() or should_skip_tracing_for_op(op):
        res = func(*args, **kwargs)
        call.output = res
        return res, call

    if _should_sample_traces(op):
        with tracing_disabled():
            res = func(*args, **kwargs)
            call.output = res
            return res, call

    # Proceed with tracing. Note that we don't check the sample rate here.
    # Only root calls get sampling applied.
    # If the parent was traced (sampled in), the child will be too.
    try:
        call = _create_call(op, *args, __weave=__weave, **kwargs)
    except OpCallError as e:
        raise e
    except Exception as e:
        if get_raise_on_captured_errors():
            raise
        log_once(
            logger.error,
            CALL_CREATE_MSG.format(traceback.format_exc()),
        )
        res = func(*args, **kwargs)
        return res, call

    # Execute the op and process the result
    client = weave_client_context.require_weave_client()
    has_finished = False

    def finish(output: Any = None, exception: BaseException | None = None) -> None:
        nonlocal has_finished
        if has_finished:
            raise ValueError("Should not call finish more than once")

        client.finish_call(
            call,
            output,
            exception,
            op=op,
        )

    def on_output(output: Any) -> Any:
        if handler := getattr(op, "_on_output_handler", None):
            return handler(output, finish, call.inputs)
        finish(output)
        return output

    try:
        res = func(*args, **kwargs)
    except Exception as e:
        finish(exception=e)
        if __should_raise:
            raise
        return None, call

    res = box.box(res)
    try:
        # Here we do a try/catch because we don't want to
        # break the user process if we trip up on processing
        # the output
        res = on_output(res)
    except Exception:
        if get_raise_on_captured_errors():
            raise
        log_once(logger.error, ON_OUTPUT_MSG.format(traceback.format_exc()))
    finally:
        # Is there a better place for this? We want to ensure that even
        # if the final output fails to be captured, we still pop the call
        # so we don't put future calls under the old call.
        call_context.pop_call(call.id)

    return res, call


async def _call_async_func(
    op: Op,
    *args: Any,
    __weave: WeaveKwargs | None = None,
    __should_raise: bool = False,
    **kwargs: Any,
) -> tuple[Any, Call]:
    func = op.resolve_fn
    call = placeholder_call()

    # Handle all of the possible cases where we would skip tracing.
    if is_tracing_setting_disabled() or should_skip_tracing_for_op(op):
        res = await func(*args, **kwargs)
        call.output = res
        return res, call

    if _should_sample_traces(op):
        with tracing_disabled():
            res = await func(*args, **kwargs)
            call.output = res
            return res, call

    # Proceed with tracing
    try:
        call = _create_call(op, *args, __weave=__weave, **kwargs)
    except OpCallError as e:
        raise e
    except Exception as e:
        if get_raise_on_captured_errors():
            raise
        log_once(
            logger.error,
            ASYNC_CALL_CREATE_MSG.format(traceback.format_exc()),
        )
        res = await func(*args, **kwargs)
        return res, call

    # Execute the op and process the result
    client = weave_client_context.require_weave_client()
    has_finished = False

    def finish(output: Any = None, exception: BaseException | None = None) -> None:
        nonlocal has_finished
        if has_finished:
            raise ValueError("Should not call finish more than once")

        client.finish_call(
            call,
            output,
            exception,
            op=op,
        )

    def on_output(output: Any) -> Any:
        if handler := getattr(op, "_on_output_handler", None):
            return handler(output, finish, call.inputs)
        finish(output)
        return output

    try:
        res = await func(*args, **kwargs)
    except Exception as e:
        finish(exception=e)
        if __should_raise:
            raise
        return None, call

    res = box.box(res)
    try:
        # Here we do a try/catch because we don't want to
        # break the user process if we trip up on processing
        # the output
        res = on_output(res)
    except Exception:
        if get_raise_on_captured_errors():
            raise
        log_once(logger.error, ON_OUTPUT_MSG.format(traceback.format_exc()))
    finally:
        # Is there a better place for this? We want to ensure that even
        # if the final output fails to be captured, we still pop the call
        # so we don't put future calls under the old call.
        call_context.pop_call(call.id)

    return res, call


def call(
    op: Op,
    *args: Any,
    __weave: WeaveKwargs | None = None,
    __should_raise: bool = False,
    **kwargs: Any,
) -> tuple[Any, Call] | Coroutine[Any, Any, tuple[Any, Call]]:
    """
    Executes the op and returns both the result and a Call representing the execution.

    This function will never raise.  Any errors are captured in the Call object.

    This method is automatically bound to any function decorated with `@weave.op`,
    allowing for usage like:

    ```python
    @weave.op
    def add(a: int, b: int) -> int:
        return a + b

    result, call = add.call(1, 2)
    ```
    """
    if inspect.iscoroutinefunction(op.resolve_fn):
        return _call_async_func(
            op,
            *args,
            __weave=__weave,
            __should_raise=__should_raise,
            **kwargs,
        )
    else:
        return _call_sync_func(
            op,
            *args,
            __weave=__weave,
            __should_raise=__should_raise,
            **kwargs,
        )


def calls(op: Op) -> CallsIter:
    """
    Get an iterator over all calls to this op.

    This method is automatically bound to any function decorated with `@weave.op`,
    allowing for usage like:

    ```python
    @weave.op
    def add(a: int, b: int) -> int:
        return a + b

    calls = add.calls()
    for call in calls:
        print(call)
    ```
    """
    client = weave_client_context.require_weave_client()
    return client._op_calls(op)


CallDisplayNameFunc = Callable[["Call"], str]
PostprocessInputsFunc = Callable[[dict[str, Any]], dict[str, Any]]
PostprocessOutputFunc = Callable[..., Any]


@overload
def op(
    func: Callable,
    *,
    name: str | None = None,
    call_display_name: str | CallDisplayNameFunc | None = None,
    postprocess_inputs: PostprocessInputsFunc | None = None,
    postprocess_output: PostprocessOutputFunc | None = None,
) -> Op: ...


@overload
def op(
    *,
    name: str | None = None,
    call_display_name: str | CallDisplayNameFunc | None = None,
    postprocess_inputs: PostprocessInputsFunc | None = None,
    postprocess_output: PostprocessOutputFunc | None = None,
) -> Callable[[Callable], Op]: ...


def op(
    func: Callable | None = None,
    *,
    name: str | None = None,
    call_display_name: str | CallDisplayNameFunc | None = None,
    postprocess_inputs: PostprocessInputsFunc | None = None,
    postprocess_output: PostprocessOutputFunc | None = None,
    tracing_sample_rate: float = 1.0,
) -> Callable[[Callable], Op] | Op:
    """
    A decorator to weave op-ify a function or method. Works for both sync and async.
    Automatically detects iterator functions and applies appropriate behavior.
    """
    if not isinstance(tracing_sample_rate, (int, float)):
        raise TypeError("tracing_sample_rate must be a float")
    if not 0 <= tracing_sample_rate <= 1:
        raise ValueError("tracing_sample_rate must be between 0 and 1")

    def op_deco(func: Callable) -> Op:
        # Check function type
        is_method = _is_unbound_method(func)
        is_async = inspect.iscoroutinefunction(func)
        is_generator = inspect.isgeneratorfunction(func)
        is_async_generator = inspect.isasyncgenfunction(func)

        # Create the appropriate wrapper based on function type
        def create_wrapper(func: Callable) -> Op:
            if is_async:

                @wraps(func)
                async def wrapper(*args: Any, **kwargs: Any) -> Any:  # pyright: ignore[reportRedeclaration]
                    res, _ = await _call_async_func(
                        cast(Op, wrapper), *args, __should_raise=True, **kwargs
                    )
                    return res
            else:

                @wraps(func)
                def wrapper(*args: Any, **kwargs: Any) -> Any:
                    res, _ = _call_sync_func(
                        cast(Op, wrapper), *args, __should_raise=True, **kwargs
                    )
                    return res

            # Tack these helpers on to our wrapper
            wrapper.resolve_fn = func  # type: ignore

            inferred_name = func.__qualname__ if is_method else func.__name__

            # funcs and methods defined inside another func will have the
            # name prefixed with {outer}.<locals>.{func_name}
            # this is noisy for us, so we strip it out
            inferred_name = inferred_name.split(".<locals>.")[-1]

            wrapper.name = name or inferred_name  # type: ignore
            wrapper.ref = None  # type: ignore

            wrapper.postprocess_inputs = postprocess_inputs  # type: ignore
            wrapper.postprocess_output = postprocess_output  # type: ignore

            wrapper.call = partial(call, wrapper)  # type: ignore
            wrapper.calls = partial(calls, wrapper)  # type: ignore

            wrapper.__call__ = wrapper  # type: ignore
            wrapper.__self__ = wrapper  # type: ignore

            wrapper._set_on_input_handler = partial(_set_on_input_handler, wrapper)  # type: ignore
            wrapper._on_input_handler = None  # type: ignore

            wrapper._set_on_output_handler = partial(_set_on_output_handler, wrapper)  # type: ignore
            wrapper._on_output_handler = None  # type: ignore

            wrapper._set_on_finish_handler = partial(_set_on_finish_handler, wrapper)  # type: ignore
            wrapper._on_finish_handler = None  # type: ignore

            wrapper._tracing_enabled = True  # type: ignore
            wrapper.tracing_sample_rate = tracing_sample_rate  # type: ignore

            wrapper.get_captured_code = partial(get_captured_code, wrapper)  # type: ignore

            if callable(call_display_name):
                params = inspect.signature(call_display_name).parameters
                if len(params) != 1:
                    raise DisplayNameFuncError(
                        "`call_display_name` function must take exactly 1 argument (the Call object)"
                    )
            wrapper.call_display_name = call_display_name  # type: ignore

            # Mark what type of function this is for runtime type checking
            wrapper._is_async = is_async  # type: ignore
            wrapper._is_generator = is_generator  # type: ignore
            wrapper._is_async_generator = is_async_generator  # type: ignore

            return cast(Op, wrapper)

        # Create the wrapper
        return create_wrapper(func)

    if func is None:
        return op_deco
    return op_deco(func)


def get_captured_code(op: Op) -> str:
    """Get the captured code of the op.

    This only works when you get an op back from a ref.  The pattern is:

    ref = weave.publish(func)
    op = ref.get()
    captured_code = op.get_captured_code()
    """
    try:
        return op.art.path_contents["obj.py"].decode()  # type: ignore
    except Exception:
        raise RuntimeError(
            "Failed to get captured code for op (this only works when you get an op back from a ref)."
        )


def maybe_bind_method(func: Callable, self: Any = None) -> Callable | MethodType:
    """Bind a function to any object (even if it's not a class)

    If self is None, return the function as is.
    """
    if (sig := inspect.signature(func)) and sig.parameters.get("self"):
        if inspect.ismethod(func) and id(func.__self__) != id(self):
            raise ValueError("Cannot re-bind a method to an new object")
        return MethodType(func, self)
    return func


def maybe_unbind_method(oplike: Op | MethodType | partial) -> Op:
    """Unbind an Op-like method or partial to a plain Op function.

    For:
    - methods, remove set `self` param
    - partials, remove any preset params
    """
    if isinstance(oplike, MethodType):
        op = oplike.__func__
    elif isinstance(oplike, partial):  # Handle cases op is defined as
        op = oplike.func
    else:
        op = oplike

    return cast(Op, op)


def is_op(obj: Any) -> bool:
    if sys.version_info < (3, 12):
        return isinstance(obj, Op)

    return all(hasattr(obj, attr) for attr in Op.__annotations__)


def as_op(fn: Callable) -> Op:
    """Given a @weave.op() decorated function, return its Op.

    @weave.op() decorated functions are instances of Op already, so this
    function should be a no-op at runtime. But you can use it to satisfy type checkers
    if you need to access OpDef attributes in a typesafe way.

    Args:
        fn: A weave.op() decorated function.

    Returns:
        The Op of the function.
    """
    if not is_op(fn):
        raise ValueError("fn must be a weave.op() decorated function")

    # The unbinding is necessary for methods because `MethodType` is applied after the
    # func is decorated into an Op.
    return maybe_unbind_method(cast(Op, fn))


_OnYieldType = Callable[[V], None]
_OnErrorType = Callable[[Exception], None]
_OnCloseType = Callable[[], None]
ON_CLOSE_MSG = "Error closing iterator, call data may be incomplete:\n{}"
ON_ERROR_MSG = "Error capturing error from iterator, call data may be incomplete:\n{}"
ON_YIELD_MSG = "Error capturing value from iterator, call data may be incomplete:\n{}"
ON_AYIELD_MSG = (
    "Error capturing async value from iterator, call data may be incomplete:\n{}"
)


class _IteratorWrapper(Generic[V]):
    """This class wraps an iterator object allowing hooks to be added to the lifecycle of the iterator. It is likely
    that this class will be helpful in other contexts and might be moved to a more general location in the future.
    """

    def __init__(
        self,
        iterator_or_ctx_manager: Iterator | AsyncIterator,
        on_yield: _OnYieldType,
        on_error: _OnErrorType,
        on_close: _OnCloseType,
    ) -> None:
        self._iterator_or_ctx_manager = iterator_or_ctx_manager
        self._on_yield = on_yield
        self._on_error = on_error
        self._on_close = on_close
        self._on_finished_called = False

        atexit.register(weakref.WeakMethod(self._call_on_close_once))

    def _call_on_close_once(self) -> None:
        if not self._on_finished_called:
            try:
                self._on_close()  # type: ignore
            except Exception as e:
                if get_raise_on_captured_errors():
                    raise
                log_once(logger.error, ON_CLOSE_MSG.format(traceback.format_exc()))
            self._on_finished_called = True

    def _call_on_error_once(self, e: Exception) -> None:
        if not self._on_finished_called:
            try:
                self._on_error(e)
            except Exception as e:
                if get_raise_on_captured_errors():
                    raise
                log_once(logger.error, ON_ERROR_MSG.format(traceback.format_exc()))
            self._on_finished_called = True

    def __iter__(self) -> _IteratorWrapper:
        return self

    def __next__(self) -> Generator[None, None, V]:
        if not hasattr(self._iterator_or_ctx_manager, "__next__"):
            try:
                # This is kept as a type ignore because the `google-generativeai` pkg seems
                # to yield an object that has properties of both value and iterator, but doesn't
                # seem to pass the isinstance(obj, Iterator) check...
                self._iterator_or_ctx_manager = iter(self._iterator_or_ctx_manager)  # type: ignore
            except TypeError:
                raise TypeError(
                    f"Cannot call next on an object of type {type(self._iterator_or_ctx_manager)}"
                )
        try:
            value = next(self._iterator_or_ctx_manager)  # type: ignore
            try:
                # Here we do a try/catch because we don't want to
                # break the user process if we trip up on processing
                # the yielded value
                self._on_yield(value)
            except Exception as e:
                # We actually use StopIteration to signal the end of the iterator
                # in some cases (like when we don't want to surface the last chunk
                # with usage info from openai integration).
                if isinstance(e, (StopAsyncIteration, StopIteration)):
                    raise
                if get_raise_on_captured_errors():
                    raise
                log_once(logger.error, ON_YIELD_MSG.format(traceback.format_exc()))
        except (StopIteration, StopAsyncIteration) as e:
            self._call_on_close_once()
            raise
        except Exception as e:
            self._call_on_error_once(e)
            raise
        else:
            return value

    def __aiter__(self) -> _IteratorWrapper:
        return self

    async def __anext__(self) -> Generator[None, None, V]:
        if not hasattr(self._iterator_or_ctx_manager, "__anext__"):
            try:
                # This is kept as a type ignore because the `google-generativeai` pkg seems
                # to yield an object that has properties of both value and iterator, but doesn't
                # seem to pass the isinstance(obj, Iterator) check...
                self._iterator_or_ctx_manager = aiter(self._iterator_or_ctx_manager)  # type: ignore
            except TypeError:
                raise TypeError(
                    f"Cannot call anext on an object of type {type(self._iterator_or_ctx_manager)}"
                )
        try:
            value = await self._iterator_or_ctx_manager.__anext__()  # type: ignore
            try:
                self._on_yield(value)
                # Here we do a try/catch because we don't want to
                # break the user process if we trip up on processing
                # the yielded value
            except Exception as e:
                # We actually use StopIteration to signal the end of the iterator
                # in some cases (like when we don't want to surface the last chunk
                # with usage info from openai integration).
                if isinstance(e, (StopAsyncIteration, StopIteration)):
                    raise
                if get_raise_on_captured_errors():
                    raise
                log_once(logger.error, ON_AYIELD_MSG.format(traceback.format_exc()))
        except (StopAsyncIteration, StopIteration) as e:
            self._call_on_close_once()
            raise StopAsyncIteration
        except Exception as e:
            self._call_on_error_once(e)
            raise
        else:
            return value

    def __del__(self) -> None:
        self._call_on_close_once()

    def close(self) -> None:
        self._call_on_close_once()

    def __getattr__(self, name: str) -> Any:
        """Delegate all other attributes to the wrapped iterator."""
        if name in [
            "_iterator_or_ctx_manager",
            "_on_yield",
            "_on_error",
            "_on_close",
            "_on_finished_called",
            "_call_on_error_once",
        ]:
            return object.__getattribute__(self, name)
        return getattr(self._iterator_or_ctx_manager, name)

    def __enter__(self) -> _IteratorWrapper:
        if hasattr(self._iterator_or_ctx_manager, "__enter__"):
            # let's enter the context manager to get the stream iterator
            self._iterator_or_ctx_manager = self._iterator_or_ctx_manager.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Exception | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        if exc_type and isinstance(exc_value, Exception):
            self._call_on_error_once(exc_value)
        if hasattr(
            self._iterator_or_ctx_manager, "__exit__"
        ):  # case where is a context mngr
            self._iterator_or_ctx_manager.__exit__(exc_type, exc_value, traceback)
        self._call_on_close_once()

    async def __aenter__(self) -> _IteratorWrapper:
        if hasattr(
            self._iterator_or_ctx_manager, "__aenter__"
        ):  # let's enter the context manager
            self._iterator_or_ctx_manager = (
                await self._iterator_or_ctx_manager.__aenter__()
            )
        return self

    async def __aexit__(
        self,
        exc_type: Exception | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        if exc_type and isinstance(exc_value, Exception):
            self._call_on_error_once(exc_value)
        self._call_on_close_once()


class _Accumulator(Generic[S, V]):
    state: S | None

    def __init__(
        self,
        accumulator: Callable[[S | None, V], S],
        initial_state: S | None = None,
    ):
        self._accumulator = accumulator
        self._state = initial_state

    def next(self, value: V) -> None:
        # the try-except hack to catch `StopIteration` inside `<integration>_accumulator`
        # this `StopIteration` is raised when some condition is met, for example, when
        # we don't want to surface last chunk (with usage info) from openai integration.
        try:
            self._state = self._accumulator(self._state, value)
        except StopIteration as e:
            self._state = e.value
            raise

    def get_state(self) -> S | None:
        return self._state


def _build_iterator_from_accumulator_for_op(
    value: Iterator[V],
    accumulator: Callable,
    on_finish: FinishCallbackType,
    iterator_wrapper: type[_IteratorWrapper] = _IteratorWrapper,
) -> _IteratorWrapper:
    acc: _Accumulator = _Accumulator(accumulator)

    def on_yield(value: V) -> None:
        acc.next(value)

    def on_error(e: Exception) -> None:
        on_finish(acc.get_state(), e)

    def on_close() -> None:
        on_finish(acc.get_state(), None)

    return iterator_wrapper(value, on_yield, on_error, on_close)


def _add_accumulator(
    op: Op,
    make_accumulator: Callable[[dict], Callable[[S, V], S]],
    *,
    should_accumulate: Callable[[dict], bool] | None = None,
    on_finish_post_processor: Callable[[Any], Any] | None = None,
    iterator_wrapper: type[_IteratorWrapper] = _IteratorWrapper,
) -> Op:
    """This is to be used internally only - specifically designed for integrations with streaming libraries.

    Add an accumulator to an op. The accumulator will be called with the output of the op
    after the op is resolved. The accumulator should return the output of the op. This is intended
    for internal use only and may change in the future. The accumulator should take two arguments:
    the current state of the accumulator and the value to accumulate. It should return the new state
    of the accumulator. The first time the accumulator is called, the current state will be None.

    The intended usage is:

    ```
    @weave.op()
    def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size

    def simple_list_accumulator(acc, value):
        if acc is None:
            acc = []
        acc.append(value)
        return acc
    add_accumulator(fn, simple_list_accumulator) # returns the op with `list(range(9, -1, -1))` as output
    """

    def on_output(
        value: Iterator[V], on_finish: FinishCallbackType, inputs: dict
    ) -> Iterator:
        def wrapped_on_finish(value: Any, e: BaseException | None = None) -> None:
            if on_finish_post_processor is not None:
                value = on_finish_post_processor(value)
            on_finish(value, e)

        if should_accumulate is None or should_accumulate(inputs):
            # we build the accumulator here dependent on the inputs (optional)
            accumulator = make_accumulator(inputs)
            return _build_iterator_from_accumulator_for_op(
                value,
                accumulator,
                wrapped_on_finish,
                iterator_wrapper,
            )
        else:
            wrapped_on_finish(value)
            return value

    op._set_on_output_handler(on_output)
    return op


__docspec__ = [call, calls]



from __future__ import annotations

import contextlib
import os
import tempfile
from collections.abc import Generator, Iterator, Mapping
from io import BytesIO, StringIO

from weave.trace.serialization import (
    op_type,  # noqa: F401, Must import this to register op save/load
)

# This uses the older weave query_service's Artifact interface. We could
# probably simplify a lot at this point by removing the internal requirement
# to use this interface.


class MemTraceFilesArtifact:
    temp_read_dir: tempfile.TemporaryDirectory | None
    path_contents: dict[str, bytes]

    def __init__(
        self,
        path_contents: Mapping[str, str | bytes] | None = None,
        metadata: dict[str, str] | None = None,
    ):
        if path_contents is None:
            path_contents = {}
        self.path_contents = path_contents  # type: ignore
        if metadata is None:
            metadata = {}
        self._metadata = metadata
        self.temp_read_dir = None

    @contextlib.contextmanager
    def new_file(self, path: str, binary: bool = False) -> Iterator[StringIO | BytesIO]:
        f: StringIO | BytesIO
        if binary:
            f = BytesIO()
        else:
            f = StringIO()
        yield f
        self.path_contents[path] = f.getvalue()  # type: ignore
        f.close()

    @property
    def is_saved(self) -> bool:
        return True

    @contextlib.contextmanager
    def open(self, path: str, binary: bool = False) -> Iterator[StringIO | BytesIO]:
        f: StringIO | BytesIO
        try:
            if binary:
                val = self.path_contents[path]
                if not isinstance(val, bytes):
                    raise ValueError(
                        f"Expected binary file, but got string for path {path}"
                    )
                f = BytesIO(val)
            else:
                val = self.path_contents[path]
                f = StringIO(val.decode("utf-8"))
        except KeyError:
            raise FileNotFoundError(path)
        yield f
        f.close()

    def path(self, path: str) -> str:
        if path not in self.path_contents:
            raise FileNotFoundError(path)

        self.temp_read_dir = tempfile.TemporaryDirectory()
        write_path = os.path.join(self.temp_read_dir.name, path)
        with open(write_path, "wb") as f:
            f.write(self.path_contents[path])
            f.flush()
            os.fsync(f.fileno())
        return write_path

    # @property
    # def metadata(self) -> artifact_fs.ArtifactMetadata:
    #     return artifact_fs.ArtifactMetadata(self._metadata, {**self._metadata})

    @contextlib.contextmanager
    def writeable_file_path(self, path: str) -> Generator[str, None, None]:
        with tempfile.TemporaryDirectory() as tmpdir:
            full_path = os.path.join(tmpdir, path)
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            yield full_path
            with open(full_path, "rb") as fp:
                self.path_contents[path] = fp.read()



"""A simple component for tabular structured data, useful as an intermediate representation."""

from __future__ import annotations

import csv
import locale
import os
import sys
import termios
import tty
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal

from rich.console import Console
from rich.table import Table

from weave.trace import util
from weave.trace.rich import pydantic_util

if TYPE_CHECKING:
    import pandas as pd


# Set locale to user's default setting
locale.setlocale(locale.LC_ALL, "")

# TODO: Add support for other types, e.g. float, datetime, etc.
ColumnType = Literal["str", "int", "bool"]
ColumnValues = list[Any]
RowValues = list[Any]


@dataclass
class ColumnDefinition:
    id: str
    label: str | None = None
    type: ColumnType | None = None
    nullable: bool = True

    # TODO: Could add support for a default value.

    @property
    def display_name(self) -> str:
        return self.label or self.id


def display_value(value: Any) -> str:
    """String representation of a grid cell value for display."""
    if value is None:
        if util.is_notebook():
            # The unicode "Symbol For Null" messes up the column lines in notebooks
            return "[gray35]<None>[/]"
        return "[gray35]␀[/]"
    if isinstance(value, bool):
        # TODO: Should we use emoji like ✅ or ❌?
        return "[green]True[/]" if value else "[red]False[/]"
    if isinstance(value, int):
        return locale.format_string("%d", value, grouping=True)
    return str(value)


def get_terminal_height() -> int:
    """Returns the number of lines available in the terminal."""
    return os.get_terminal_size().lines - 5  # Adjust for headers & padding


def get_key() -> str:
    """Reads a single keypress from the user without requiring Enter."""
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(fd)
        ch = sys.stdin.read(1)
        # Check if it might be an arrow key (starts with escape sequence)
        if ch == "\x1b":
            # Read the next two characters for arrow keys
            next_chars = sys.stdin.read(2)
            return ch + next_chars
        else:
            return ch
    except KeyboardInterrupt:
        return "q"  # Handle Ctrl+C gracefully
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)


class Row:
    """Access to one row of a Grid."""

    grid: Grid
    index: int

    def __init__(self, grid: Grid, index: int) -> None:
        self.grid = grid
        self.index = index

    def to_list(self) -> RowValues:
        return self.grid.rows[self.index]

    def to_dict(self) -> dict[str, Any]:
        return {col.id: self[col.id] for col in self.grid.columns}

    def to_rich_table(self) -> Table:
        """Convert the row to a Rich table with key-value pairs."""
        d = {col.display_name: display_value(self[col.id]) for col in self.grid.columns}
        return pydantic_util.dict_to_table(d)

    def to_rich_table_str(self) -> Table:
        table = self.to_rich_table()
        return pydantic_util.table_to_str(table)

    def __getattr__(self, name: str) -> Any:
        """Allow accessing column values as attributes.

        Example:
            row.column_id instead of row['column_id']

        Args:
            name: The column id to access

        Returns:
            The value in the specified column

        Raises:
            AttributeError: If the column doesn't exist
        """
        try:
            return self[name]
        except KeyError:
            raise AttributeError(f"Row has no attribute or column '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        """Allow setting column values as attributes.

        Example:
            row.column_id = value instead of row['column_id'] = value

        Args:
            name: The column id to set
            value: The value to assign to the column
        """
        if name in ("grid", "index"):
            # These are actual attributes of the Row class
            object.__setattr__(self, name, value)
        else:
            try:
                self[name] = value
            except KeyError:
                object.__setattr__(self, name, value)

    def __getitem__(self, key: str | int) -> Any:
        if isinstance(key, str):
            col_index = self.grid._column_id_to_index.get(key)
            if col_index is None:
                raise KeyError(f"Column '{key}' not found")
            return self.grid.rows[self.index][col_index]
        elif isinstance(key, int):
            if key < 0 or key >= len(self.grid.rows[self.index]):
                raise IndexError("Row index out of range")
            return self.grid.rows[self.index][key]
        else:
            raise TypeError("Key must be string or integer")

    def __setitem__(self, key: str | int, value: Any) -> None:
        """Set a value in the row using either column name or index.

        Args:
            key: Column id (str) or index (int)
            value: The value to set
        """
        if isinstance(key, (str, int)):
            self.grid.set_cell_value(self.index, key, value)
        else:
            raise TypeError("Key must be string or integer")

    def _repr_pretty_(self, p: Any, cycle: bool) -> None:
        """Pretty representation for IPython."""
        if cycle:
            p.text("Row(...)")
        else:
            items = [
                f"{col.id}={display_value(self[col.id])}" for col in self.grid.columns
            ]
            p.text(f"Row({', '.join(items)})")


class Column:
    grid: Grid
    column_id: str

    def __init__(self, grid: Grid, column_id: str) -> None:
        self.grid = grid
        self.column_id = column_id

    def rename(self, label: str) -> Column:
        """Rename the column."""
        self.grid.rename_column(self.column_id, label)
        return self

    def to_list(self) -> ColumnValues:
        """Convert column values to a list.

        Returns:
            A list containing all values in this column.
        """
        col_index = self.grid._column_id_to_index.get(self.column_id)
        if col_index is None:
            raise KeyError(f"Column '{self.column_id}' not found")
        return [row[col_index] for row in self.grid.rows]

    def __getitem__(self, index: int | slice) -> Any | list[Any]:
        """Get the value(s) at the specified row index or slice for this column.

        Args:
            index: The row index or slice to retrieve the value(s) from

        Returns:
            The value at the specified row index or a list of values for a slice

        Raises:
            IndexError: If the row index is out of range
        """
        col_index = self.grid._column_id_to_index.get(self.column_id)
        if col_index is None:
            raise KeyError(f"Column '{self.column_id}' not found")

        num_rows = len(self.grid.rows)
        if isinstance(index, slice):
            # Handle slice by returning a list of values
            start, stop, step = index.indices(num_rows)
            return [self.grid.rows[i][col_index] for i in range(start, stop, step)]
        else:
            # Handle single index
            if index < 0 or index >= num_rows:
                raise IndexError("Row index out of range")
            return self.grid.rows[index][col_index]

    def __setitem__(self, index: int, value: Any) -> None:
        """Set the value at the specified row index for this column.

        Args:
            index: The row index to set the value at
            value: The value to set

        Raises:
            IndexError: If the row index is out of range
        """
        if index < 0 or index >= len(self.grid.rows):
            raise IndexError("Row index out of range")

        self.grid.set_cell_value(index, self.column_id, value)

    def to_rich_table(self) -> Table:
        """Convert the column to a rich Table for pretty display.

        Returns:
            A rich Table with a single column containing all values from this column.
        """
        table = Table()

        # Add a single column with the column name
        col_def = next(
            (col for col in self.grid.columns if col.id == self.column_id), None
        )
        display_name = col_def.display_name if col_def else self.column_id

        # Set justification based on column type
        if col_def and col_def.type == "int":
            table.add_column(display_name, justify="right")
        else:
            table.add_column(display_name)

        # Add all values from this column as rows
        for i in range(len(self.grid.rows)):
            try:
                value = self[i]
                # Convert value to string for display
                table.add_row(display_value(value))
            except (IndexError, KeyError):
                table.add_row("N/A")

        return table

    def to_rich_table_str(self) -> str:
        table = self.to_rich_table()
        return pydantic_util.table_to_str(table)

    def _repr_pretty_(self, p: Any, cycle: bool) -> None:
        """Show a nicely formatted table in ipython."""
        if cycle:
            p.text("Column(...)")
        else:
            p.text(self.to_rich_table_str())


class Grid:
    columns: list[ColumnDefinition]
    rows: list[RowValues]
    _column_id_to_index: dict[str, int]

    def __init__(self, name: str | None = None) -> None:
        self.columns = []
        self.rows = []
        self._column_id_to_index = {}

    def set_cell_value(self, row_index: int, column: str | int, value: Any) -> Grid:
        """Set the value of a specific cell in the grid.

        Args:
            row_index: The index of the row to update
            column: The name or index of the column to update
            value: The new value to set

        Returns:
            The grid instance for method chaining
        """
        if row_index < 0 or row_index >= len(self.rows):
            raise IndexError("Row index out of range")

        column_index = None
        if isinstance(column, int):
            if column < 0 or column >= len(self.columns):
                raise IndexError("Column index out of range")
            column_index = column
        elif isinstance(column, str):
            for i, col in enumerate(self.columns):
                if col.id == column:
                    column_index = i
                    break
            if column_index is None:
                raise KeyError(f"Column '{column}' not found")
        else:
            raise TypeError("Column must be specified by name (str) or index (int)")

        # Check column type and coerce value if possible
        col = self.columns[column_index]
        if value is None and not col.nullable:
            raise ValueError(f"Column '{col.id}' is not nullable")
        elif col.type is not None:
            try:
                if col.type == "bool" and not isinstance(value, bool):
                    value = bool(value)
                elif col.type == "int" and not isinstance(value, int):
                    value = int(value)
                elif col.type == "str" and not isinstance(value, str):
                    value = str(value)
            except (ValueError, TypeError):
                raise ValueError(
                    f"Cannot coerce value '{value}' to type '{col.type}' for column '{col.id}'"
                )

        self.rows[row_index][column_index] = value
        return self

    def insert_column(
        self,
        column_index: int,
        name: str,
        label: str | None = None,
        type: ColumnType | None = None,
    ) -> Grid:
        """Insert a new column at the specified index in the grid."""
        self.columns.insert(
            column_index, ColumnDefinition(id=name, label=label, type=type)
        )
        self._column_id_to_index = {col.id: i for i, col in enumerate(self.columns)}
        for row_index in range(self.num_rows):
            self.rows[row_index].insert(column_index, None)
        return self

    def add_column(
        self, name: str, label: str | None = None, type: ColumnType | None = None
    ) -> Grid:
        """Add a new column to the grid."""
        return self.insert_column(len(self.columns), name, label, type)

    def _get_column_definition(self, column_id: int | str) -> ColumnDefinition:
        """Get the definition of a column by id."""
        if isinstance(column_id, int):
            if column_id < 0 or column_id >= len(self.columns):
                raise IndexError("Column index out of range")
            return self.columns[column_id]
        for col in self.columns:
            if col.id == column_id:
                return col
        raise KeyError(f"Column '{column_id}' not found")

    def rename_column(self, column_id: int | str, label: str) -> Grid:
        """Rename a column in the grid."""
        definition = self._get_column_definition(column_id)
        definition.label = label
        return self

    def add_row(self, values: RowValues) -> Grid:
        """Add a new row of values to the grid."""
        if len(values) != len(self.columns):
            raise ValueError(
                f"Row length ({len(values)}) does not match number of columns ({len(self.columns)})"
            )

        # Check column types and coerce values if possible
        processed_values: RowValues = []
        for i, (value, column) in enumerate(zip(values, self.columns)):
            if column.type is not None:
                try:
                    if column.type == "bool" and not isinstance(value, bool):
                        processed_values.append(bool(value))
                    elif column.type == "int" and not isinstance(value, int):
                        processed_values.append(int(value))
                    elif column.type == "str" and not isinstance(value, str):
                        processed_values.append(str(value))
                    else:
                        processed_values.append(value)
                except (ValueError, TypeError):
                    raise ValueError(
                        f"Cannot coerce value '{value}' to type '{column.type}' for column '{column.id}'"
                    )
            else:
                processed_values.append(value)

        self.rows.append(processed_values)
        return self

    def get_row(self, index: int) -> Row:
        """Get a row by index that allows access by column name or position."""
        if index < 0 or index >= len(self.rows):
            raise IndexError("Row index out of range")
        return Row(self, index)
        # return Row(self.rows[index], self.columns)

    def get_column_values(self, column: str | int) -> ColumnValues:
        """Get all values for a specific column by name or index."""
        if isinstance(column, str):
            # Find the column index by name
            for i, col in enumerate(self.columns):
                if col.id == column:
                    column_index = i
                    break
            else:
                raise KeyError(f"Column '{column}' not found")
        elif isinstance(column, int):
            if column < 0 or column >= len(self.columns):
                raise IndexError("Column index out of range")
            column_index = column
        else:
            raise TypeError(
                "Column must be specified by name (string) or index (integer)"
            )

        # Extract values for the specified column from all rows
        return [row[column_index] for row in self.rows]

    def get_column(self, key: str | int) -> Column:
        """Get a column by id."""
        if isinstance(key, int):
            if key < 0 or key >= len(self.columns):
                raise IndexError("Column index out of range")
            column_id = self.columns[key].id
        elif isinstance(key, str):
            if key not in self._column_id_to_index:
                raise KeyError(f"Column '{key}' not found")
            column_id = key
        else:
            raise TypeError(
                "Key must be a string (column name) or integer (column index)"
            )
        return Column(self, column_id)

    def __getitem__(self, key: str | int | slice) -> Column | Row | Grid:
        """
        Allow accessing column values using grid['column_name'] syntax,
        accessing a row using grid[row_index] syntax,
        or slicing rows using grid[start:end] syntax.
        """
        if isinstance(key, str):
            return self.get_column(key)
        elif isinstance(key, int):
            return self.get_row(key)
        elif isinstance(key, slice):
            # Create a new grid with just the sliced rows
            new_grid = Grid()
            new_grid.columns = self.columns.copy()
            new_grid._column_id_to_index = self._column_id_to_index.copy()
            start, stop, step = key.indices(len(self.rows))
            new_grid.rows = self.rows[start:stop:step]
            return new_grid
        else:
            raise TypeError(
                "Key must be a string (column name), integer (row index), or slice (row range)"
            )

    @property
    def num_rows(self) -> int:
        """Return the number of rows in the grid."""
        return len(self.rows)

    @property
    def num_columns(self) -> int:
        """Return the number of columns in the grid."""
        return len(self.columns)

    def to_rich_table(self, row_start: int = 0, row_end: int | None = None) -> Table:
        """Convert the grid spec (not calls) to a rich Table for pretty display."""
        table = Table()

        # Add columns to the table
        for col in self.columns:
            if col.type == "int":
                table.add_column(
                    col.display_name, justify="right", header_style="bold cyan"
                )
            else:
                table.add_column(col.display_name, header_style="bold cyan")

        # Add rows to the table
        for row_values in self.rows[row_start:row_end]:
            # Convert all values to strings for display
            str_values = [display_value(val) for val in row_values]
            table.add_row(*str_values)

        return table

    def _repr_pretty_(self, p: Any, cycle: bool) -> None:
        """Show a nicely formatted table in ipython."""
        if cycle:
            p.text("Grid(...)")
        else:
            p.text(self.to_rich_table_str())

    def to_rich_table_str(self) -> str:
        table = self.to_rich_table()
        return pydantic_util.table_to_str(table)

    def show(self, rows_per_page: int | None = None) -> None:
        height = get_terminal_height()
        if self.num_rows < height:
            print(self.to_rich_table_str())
            return

        height -= 2  # Adjust for pagination and navigation instructions

        if rows_per_page is None:
            # Make rows_per_page a multiple of 5 with a minimum value of 5
            rows_per_page = max(5, (height // 5) * 5)
        page = 0
        total_pages = (self.num_rows // rows_per_page) + (
            1 if self.num_rows % rows_per_page else 0
        )

        console = Console()
        while True:
            console.clear()
            console.print(f"Page {page + 1}/{total_pages}", style="bold magenta")
            paginated_table = self.to_rich_table(
                page * rows_per_page, (page + 1) * rows_per_page
            )
            console.print(paginated_table)
            console.print(
                "\nUse ← (left) / → (right) to navigate, 'q' to quit",
                style="bold yellow",
            )

            key = get_key()
            if key == "\x1b[C" and page < total_pages - 1:  # Right arrow key
                page += 1
            elif key == "\x1b[D" and page > 0:  # Left arrow key
                page -= 1
            elif key.lower() == "q":  # Quit on 'q'
                break

    # TODO: Add additional methods for converting to CSV/TSV/JSON?

    def to_pandas(self) -> pd.DataFrame:
        """Convert the grid contents to a pandas DataFrame.

        Returns:
            A pandas DataFrame containing all the data from the grid.
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("pandas is required to use this method")

        # Create a DataFrame directly from the grid data
        data: dict[str, list[Any]] = {col.id: [] for col in self.columns}

        # Populate the data dictionary with values from each row
        for row_values in self.rows:
            for i, col in enumerate(self.columns):
                data[col.id].append(row_values[i] if i < len(row_values) else None)

        return pd.DataFrame(data)

    @classmethod
    def from_pandas(cls, df: pd.DataFrame) -> Grid:
        """Create a Grid from a pandas DataFrame.

        Args:
            df: The pandas DataFrame to convert to a Grid

        Returns:
            A Grid instance containing the data from the DataFrame
        """
        grid = cls()

        # Add columns based on DataFrame columns
        for col_name in df.columns:
            # Try to determine column type
            col_type: ColumnType | None = None
            if df[col_name].dtype == "int64":
                col_type = "int"

            grid.add_column(str(col_name), type=col_type)

        # Add rows from DataFrame
        for _, row in df.iterrows():
            values = []
            for col in df.columns:
                values.append(row[col])
            grid.add_row(values)

        return grid

    @classmethod
    def load_csv(cls, file_path: str | os.PathLike, **kwargs: Any) -> Grid:
        """Create a Grid from a CSV file.

        Args:
            file_path: Path to the CSV file
            **kwargs: Additional arguments to pass to csv.reader (like delimiter, quotechar)

        Returns:
            A Grid instance containing the data from the CSV file

        Raises:
            FileNotFoundError: If the file does not exist
        """
        grid = cls()

        with open(file_path, newline="", encoding="utf-8") as csv_file:
            reader = csv.reader(csv_file, **kwargs)
            # Read header row for column names
            try:
                header = next(reader)
            except StopIteration:
                # Empty file
                return grid

            # Add columns based on header
            for col_name in header:
                grid.add_column(col_name)

            # Process data rows
            for row in reader:
                # Try to infer types for the first data row
                # TODO: Maybe should consider first N rows to infer types?
                if len(grid.rows) == 0:
                    # TODO: Ability to infer bool
                    for i, value in enumerate(row):
                        # Try to convert to int
                        try:
                            int(value)
                            # If successful, update column type to int
                            if i < len(grid.columns):
                                grid.columns[i].type = "int"
                        except (ValueError, TypeError):
                            pass

                # Add the row data
                processed_row: RowValues = []
                for i, value in enumerate(row):
                    # Convert to int if the column type is int
                    if i < len(grid.columns) and grid.columns[i].type == "int":
                        try:
                            processed_row.append(int(value) if value else None)
                        except (ValueError, TypeError):
                            processed_row.append(None)
                    else:
                        processed_row.append(value)

                # Pad row if needed
                while len(processed_row) < len(grid.columns):
                    processed_row.append(None)

                grid.add_row(processed_row)

        return grid



from typing import Callable, Optional

from pydantic import BaseModel

import weave
from weave.trace.autopatch import OpSettings
from weave.trace.op import _add_accumulator


def instructor_partial_accumulator(
    acc: Optional[BaseModel], value: BaseModel
) -> BaseModel:
    if acc is None or acc != value:
        acc = value
    return acc


def instructor_wrapper_partial(settings: OpSettings) -> Callable[[Callable], Callable]:
    def wrapper(fn: Callable) -> Callable:
        op_kwargs = settings.model_dump()
        op = weave.op(fn, **op_kwargs)
        return _add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda inputs: instructor_partial_accumulator,
            should_accumulate=lambda inputs: True,
        )

    return wrapper



from __future__ import annotations

import logging
from collections.abc import Sequence
from types import CoroutineType
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from weave.trace.object_record import ObjectRecord
from weave.trace.refs import ObjectRef, TableRef, parse_uri
from weave.trace.sanitize import REDACTED_VALUE, should_redact
from weave.trace.serialization import custom_objs
from weave.trace.serialization.dictifiable import try_to_dict
from weave.trace_server.interface.builtin_object_classes.builtin_object_registry import (
    BUILTIN_OBJECT_REGISTRY,
)
from weave.trace_server.trace_server_interface import (
    FileContentReadReq,
    FileCreateReq,
    TraceServerInterface,
)
from weave.trace_server.trace_server_interface_util import bytes_digest

if TYPE_CHECKING:
    from weave.trace.weave_client import WeaveClient


def is_pydantic_model_class(obj: Any) -> bool:
    """Determine if obj is a subclass of pydantic.BaseModel."""
    try:
        return (
            isinstance(obj, type)
            and issubclass(obj, BaseModel)
            and obj is not BaseModel
        )
    except TypeError:
        # Might be something like Iterable[CalendarEvent]
        return False


def _is_inline_custom_obj(encoded: dict) -> bool:
    """Custom object values may be inline or backed by a file.

    This separate function for readability checks which we are dealing with."""
    return "val" in encoded


def to_json(
    obj: Any, project_id: str, client: WeaveClient, use_dictify: bool = False
) -> Any:
    if isinstance(obj, TableRef):
        return obj.uri()
    elif isinstance(obj, ObjectRef):
        return obj.uri()
    elif isinstance(obj, ObjectRecord):
        res = {"_type": obj._class_name}
        for k, v in obj.__dict__.items():
            res[k] = to_json(v, project_id, client, use_dictify)
        return res
    elif isinstance_namedtuple(obj):
        return {
            k: to_json(v, project_id, client, use_dictify)
            for k, v in obj._asdict().items()
        }
    elif isinstance(obj, (list, tuple)):
        return [to_json(v, project_id, client, use_dictify) for v in obj]
    elif isinstance(obj, dict):
        return {k: to_json(v, project_id, client, use_dictify) for k, v in obj.items()}
    elif is_pydantic_model_class(obj):
        return obj.model_json_schema()

    if isinstance(obj, (int, float, str, bool)) or obj is None:
        return obj

    # Add explicit handling for WeaveScorerResult models
    from weave.flow.scorer import WeaveScorerResult

    if isinstance(obj, WeaveScorerResult):
        return {
            k: to_json(v, project_id, client, use_dictify)
            for k, v in obj.model_dump().items()
        }

    # This still blocks potentially on large-file i/o.
    encoded = custom_objs.encode_custom_obj(obj)
    if encoded is None:
        if (
            use_dictify
            and not isinstance(obj, ALWAYS_STRINGIFY)
            and not has_custom_repr(obj)
        ):
            return dictify(obj)

        # TODO: I would prefer to only have this once in dictify? Maybe dictify and to_json need to be merged?
        # However, even if dictify is false, i still want to try to convert to dict
        elif as_dict := try_to_dict(obj):
            return {
                k: to_json(v, project_id, client, use_dictify)
                for k, v in as_dict.items()
            }
        return fallback_encode(obj)
    if _is_inline_custom_obj(encoded):
        return encoded
    result = _build_result_from_encoded(encoded, project_id, client)
    return result


def _build_result_from_encoded(
    encoded: dict, project_id: str, client: WeaveClient
) -> Any:
    file_digests = {}
    for name, val in encoded["files"].items():
        # Instead of waiting for the file to be created, we
        # calculate the digest directly. This makes sure that the
        # to_json procedure is not blocked on network requests.
        # Technically it is possible that the file creation request
        # fails.
        client._send_file_create(
            FileCreateReq(project_id=project_id, name=name, content=val)
        )
        contents_as_bytes = val
        if isinstance(contents_as_bytes, str):
            contents_as_bytes = contents_as_bytes.encode("utf-8")
        digest = bytes_digest(contents_as_bytes)
        file_digests[name] = digest
    result = {
        "_type": encoded["_type"],
        "weave_type": encoded["weave_type"],
        "files": file_digests,
    }
    load_op_uri = encoded.get("load_op")
    if load_op_uri:
        result["load_op"] = load_op_uri
    return result


MAX_STR_LEN = 1000


def stringify(obj: Any, limit: int = MAX_STR_LEN) -> str:
    """This is a fallback for objects that we don't have a better way to serialize."""
    rep = None
    try:
        rep = repr(obj)
    except Exception:
        try:
            rep = str(obj)
        except Exception:
            rep = f"<{type(obj).__name__}: {id(obj)}>"
    if isinstance(rep, str):
        if len(rep) > limit:
            rep = rep[: limit - 3] + "..."
    return rep


def is_primitive(obj: Any) -> bool:
    """Check if an object is a known primitive type."""
    return isinstance(obj, (int, float, str, bool, type(None)))


def has_custom_repr(obj: Any) -> bool:
    """Return True if the object has a custom __repr__ method."""
    return obj.__class__.__repr__ is not object.__repr__


def dictify(
    obj: Any, maxdepth: int = 0, depth: int = 1, seen: set[int] | None = None
) -> Any:
    """Recursively compute a dictionary representation of an object."""
    if seen is None:
        seen = set()

    if not is_primitive(obj):
        obj_id = id(obj)
        if obj_id in seen:
            # Avoid infinite recursion with circular references
            return stringify(obj)
        else:
            seen.add(obj_id)

    if maxdepth > 0 and depth > maxdepth:
        # TODO: If obj at this point is a simple type,
        #       maybe we should just return it rather than stringify
        return stringify(obj)

    if is_primitive(obj):
        return obj
    elif isinstance(obj, (list, tuple)):
        return [dictify(v, maxdepth, depth + 1, seen) for v in obj]
    elif isinstance(obj, dict):
        dict_result = {}
        for k, v in obj.items():
            if isinstance(k, str) and should_redact(k):
                dict_result[k] = REDACTED_VALUE
            else:
                dict_result[k] = dictify(v, maxdepth, depth + 1, seen)
        return dict_result

    if hasattr(obj, "to_dict"):
        try:
            as_dict = obj.to_dict()
            if isinstance(as_dict, dict):
                to_dict_result = {}
                for k, v in as_dict.items():
                    if isinstance(k, str) and should_redact(k):
                        to_dict_result[k] = REDACTED_VALUE
                    elif maxdepth == 0 or depth < maxdepth:
                        to_dict_result[k] = dictify(v, maxdepth, depth + 1)
                    else:
                        to_dict_result[k] = stringify(v)
                return to_dict_result
        except Exception:
            raise ValueError("to_dict failed")

    result: dict[Any, Any] = {}
    result["__class__"] = {
        "module": obj.__class__.__module__,
        "qualname": obj.__class__.__qualname__,
        "name": obj.__class__.__name__,
    }
    if isinstance(obj, Sequence) and not isinstance(obj, (str, bytes)):
        # Custom list-like object
        try:
            for i, item in enumerate(obj):
                result[i] = dictify(item, maxdepth, depth + 1, seen)
        except Exception:
            return stringify(obj)
    else:
        for attr in dir(obj):
            if attr.startswith("_"):
                continue
            if should_redact(attr):
                result[attr] = REDACTED_VALUE
                continue
            try:
                val = getattr(obj, attr)
                if callable(val):
                    continue
                if maxdepth == 0 or depth < maxdepth:
                    result[attr] = dictify(val, maxdepth, depth + 1, seen)
                else:
                    result[attr] = stringify(val)
            except Exception:
                return stringify(obj)
    return result


# TODO: Investigate why dictifying Logger causes problems
ALWAYS_STRINGIFY = (logging.Logger, CoroutineType)


# Note: Max depth not picked scientifically, just trying to keep things under control.
DEFAULT_MAX_DICTIFY_DEPTH = 10


def fallback_encode(obj: Any) -> Any:
    rep = None
    try:
        rep = repr(obj)
    except Exception:
        try:
            rep = str(obj)
        except Exception:
            rep = f"<{type(obj).__name__}: {id(obj)}>"
    if isinstance(rep, str):
        if len(rep) > MAX_STR_LEN:
            rep = rep[:MAX_STR_LEN] + "..."
    return rep


def isinstance_namedtuple(obj: Any) -> bool:
    return (
        isinstance(obj, tuple) and hasattr(obj, "_asdict") and hasattr(obj, "_fields")
    )


def _load_custom_obj_files(
    project_id: str, server: TraceServerInterface, file_digests: dict
) -> dict[str, bytes]:
    loaded_files: dict[str, bytes] = {}
    for name, digest in file_digests.items():
        file_response = server.file_content_read(
            FileContentReadReq(project_id=project_id, digest=digest)
        )
        loaded_files[name] = file_response.content
    return loaded_files


def from_json(obj: Any, project_id: str, server: TraceServerInterface) -> Any:
    if isinstance(obj, list):
        return [from_json(v, project_id, server) for v in obj]
    elif isinstance(obj, dict):
        if (val_type := obj.pop("_type", None)) is None:
            return {k: from_json(v, project_id, server) for k, v in obj.items()}
        elif val_type == "ObjectRecord":
            return ObjectRecord(
                {k: from_json(v, project_id, server) for k, v in obj.items()}
            )
        elif val_type == "CustomWeaveType":
            if _is_inline_custom_obj(obj):
                return custom_objs.decode_custom_inline_obj(obj)
            files = _load_custom_obj_files(project_id, server, obj["files"])
            return custom_objs.decode_custom_files_obj(
                obj["weave_type"], files, obj.get("load_op")
            )
        elif (
            isinstance(val_type, str)
            and obj.get("_class_name") == val_type
            and (builtin_object_class := BUILTIN_OBJECT_REGISTRY.get(val_type))
        ):
            return builtin_object_class.model_validate(obj)
        else:
            return ObjectRecord(
                {k: from_json(v, project_id, server) for k, v in obj.items()}
            )
    elif isinstance(obj, str) and obj.startswith("weave://"):
        return parse_uri(obj)

    return obj



from pydantic import BaseModel

from weave.trace_server.interface.builtin_object_classes import base_object_def


class TestOnlyNestedBaseModel(BaseModel):
    a: int


class TestOnlyNestedBaseObject(base_object_def.BaseObject):
    b: int


class TestOnlyExample(base_object_def.BaseObject):
    primitive: int
    nested_base_model: TestOnlyNestedBaseModel
    # Important: `RefStr` is just an alias for `str`. When defining `BaseObject`s, we
    # should never have a property point to another `BaseObject`. This is because each
    # base object is stored in the database and should be treated like a foreign key.
    #
    # It would be nice to have a way to ensure that no `BaseObject` has any `BaseObject`
    # properties.
    nested_base_object: base_object_def.RefStr


__all__ = ["TestOnlyExample", "TestOnlyNestedBaseObject"]



from typing import TYPE_CHECKING, Any, Callable, Optional

import weave
from weave.trace.op import _add_accumulator, _IteratorWrapper
from weave.trace.weave_client import Call

if TYPE_CHECKING:
    from botocore.client import BaseClient


def bedrock_on_finish_converse(
    call: Call, output: Any, exception: Optional[BaseException]
) -> None:
    model_name = str(call.inputs["modelId"])  # get the ref
    usage = {model_name: {"requests": 1}}
    summary_update = {"usage": usage}
    if output:
        tokens_metrics = {
            "prompt_tokens": output["usage"]["inputTokens"],
            "completion_tokens": output["usage"]["outputTokens"],
            "total_tokens": output["usage"]["totalTokens"],
        }
        usage[model_name].update(tokens_metrics)
    if call.summary is not None:
        call.summary.update(summary_update)


def bedrock_on_finish_invoke(
    call: Call, output: Any, exception: Optional[BaseException]
) -> None:
    model_name = str(call.inputs["modelId"])
    usage = {model_name: {"requests": 1}}
    summary_update = {"usage": usage}

    if output and "ResponseMetadata" in output:
        headers = output["ResponseMetadata"]["HTTPHeaders"]
        prompt_tokens = int(headers.get("x-amzn-bedrock-input-token-count", 0))
        completion_tokens = int(headers.get("x-amzn-bedrock-output-token-count", 0))
        total_tokens = prompt_tokens + completion_tokens

        tokens_metrics = {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": completion_tokens,
            "total_tokens": total_tokens,
        }
        usage[model_name].update(tokens_metrics)
    if call.summary is not None:
        call.summary.update(summary_update)


def postprocess_inputs(inputs: dict[str, Any]) -> dict[str, Any]:
    return inputs.get("kwargs", {})


def _patch_converse(bedrock_client: "BaseClient") -> None:
    op = weave.op(
        bedrock_client.converse,
        name="BedrockRuntime.converse",
        postprocess_inputs=postprocess_inputs,
    )
    op._set_on_finish_handler(bedrock_on_finish_converse)
    bedrock_client.converse = op


def _patch_invoke(bedrock_client: "BaseClient") -> None:
    op = weave.op(
        bedrock_client.invoke_model,
        name="BedrockRuntime.invoke",
        postprocess_inputs=postprocess_inputs,
    )
    op._set_on_finish_handler(bedrock_on_finish_invoke)
    bedrock_client.invoke_model = op


def bedrock_stream_accumulator(
    acc: Optional[dict],
    value: dict,
) -> dict:
    """Accumulates streaming events into a final response dictionary."""
    if acc is None:
        acc = {
            "role": None,
            "content": "",
            "stop_reason": None,
            "usage": {
                "inputTokens": 0,
                "outputTokens": 0,
                "totalTokens": 0,
            },
            "latency_ms": None,
        }

    # Handle 'messageStart' event
    if "messageStart" in value:
        acc["role"] = value["messageStart"]["role"]

    # Handle 'contentBlockDelta' event
    if "contentBlockDelta" in value:
        acc["content"] += value["contentBlockDelta"]["delta"]["text"]

    # Handle 'messageStop' event
    if "messageStop" in value:
        acc["stop_reason"] = value["messageStop"]["stopReason"]

    # Handle 'metadata' event
    if "metadata" in value:
        metadata = value["metadata"]
        if "usage" in metadata:
            acc["usage"]["inputTokens"] = metadata["usage"].get("inputTokens", 0)
            acc["usage"]["outputTokens"] = metadata["usage"].get("outputTokens", 0)
            acc["usage"]["totalTokens"] = metadata["usage"].get("totalTokens", 0)
        if "metrics" in metadata:
            acc["latency_ms"] = metadata["metrics"].get("latencyMs", 0)

    return acc


def create_stream_wrapper(
    name: str,
) -> Callable[[Callable], Callable]:
    def wrapper(fn: Callable) -> Callable:
        op = weave.op(postprocess_inputs=postprocess_inputs)(fn)
        op.name = name  # type: ignore
        op._set_on_finish_handler(bedrock_on_finish_converse)

        class BedrockIteratorWrapper(_IteratorWrapper):
            def get(self, key: str, default: Any = None) -> Any:
                """Delegate 'get' method to the response object."""
                if key == "stream":
                    if hasattr(self._iterator_or_ctx_manager, "get"):
                        self._iterator_or_ctx_manager = (
                            self._iterator_or_ctx_manager.get("stream")
                        )
                    return self

        return _add_accumulator(
            op,
            make_accumulator=lambda _: bedrock_stream_accumulator,
            should_accumulate=lambda _: True,
            iterator_wrapper=BedrockIteratorWrapper,
        )

    return wrapper


def _patch_converse_stream(bedrock_client: "BaseClient") -> None:
    """Patches the converse_stream method to handle streaming."""
    op = create_stream_wrapper("BedrockRuntime.converse_stream")(
        bedrock_client.converse_stream
    )
    bedrock_client.converse_stream = op


def patch_client(bedrock_client: "BaseClient") -> None:
    _patch_converse(bedrock_client)
    _patch_invoke(bedrock_client)
    _patch_converse_stream(bedrock_client)



from __future__ import annotations

import importlib
from typing import TYPE_CHECKING, Callable

import weave
from weave.integrations.patcher import MultiPatcher, NoOpPatcher, SymbolPatcher
from weave.trace.autopatch import IntegrationSettings, OpSettings
from weave.trace.op import _add_accumulator

if TYPE_CHECKING:
    from mistralai.models.chat_completion import (
        ChatCompletionResponse,
        ChatCompletionStreamResponse,
    )

_mistral_patcher: MultiPatcher | None = None


def mistral_accumulator(
    acc: ChatCompletionResponse | None,
    value: ChatCompletionStreamResponse,
) -> ChatCompletionResponse:
    # This import should be safe at this point
    from mistralai.models.chat_completion import (
        ChatCompletionResponse,
        ChatCompletionResponseChoice,
        ChatMessage,
    )
    from mistralai.models.common import UsageInfo

    if acc is None:
        acc = ChatCompletionResponse(
            id=value.id,
            object=value.object,
            created=value.created,
            model=value.model,
            choices=[],
            usage=UsageInfo(prompt_tokens=0, total_tokens=0, completion_tokens=None),
        )

    # Merge in the usage info
    if value.usage is not None:
        acc.usage.prompt_tokens += value.usage.prompt_tokens
        acc.usage.total_tokens += value.usage.total_tokens
        if acc.usage.completion_tokens is None:
            acc.usage.completion_tokens = value.usage.completion_tokens
        else:
            acc.usage.completion_tokens += value.usage.completion_tokens

    # Loop through the choices and add their deltas
    for delta_choice in value.choices:
        while delta_choice.index >= len(acc.choices):
            acc.choices.append(
                ChatCompletionResponseChoice(
                    index=len(acc.choices),
                    message=ChatMessage(role="", content=""),
                    finish_reason=None,
                )
            )
        acc.choices[delta_choice.index].message.role = (
            delta_choice.delta.role or acc.choices[delta_choice.index].message.role
        )
        acc.choices[delta_choice.index].message.content += (
            delta_choice.delta.content or ""
        )
        if delta_choice.delta.tool_calls:
            if acc.choices[delta_choice.index].message.tool_calls is None:
                acc.choices[delta_choice.index].message.tool_calls = []
            acc.choices[
                delta_choice.index
            ].message.tool_calls += delta_choice.delta.tool_calls
        acc.choices[delta_choice.index].finish_reason = (
            delta_choice.finish_reason or acc.choices[delta_choice.index].finish_reason
        )

    return acc


def mistral_stream_wrapper(settings: OpSettings) -> Callable:
    def wrapper(fn: Callable) -> Callable:
        op_kwargs = settings.model_dump()
        op = weave.op(fn, **op_kwargs)
        acc_op = _add_accumulator(op, lambda inputs: mistral_accumulator)  # type: ignore
        return acc_op

    return wrapper


def mistral_wrapper(settings: OpSettings) -> Callable:
    def wrapper(fn: Callable) -> Callable:
        op_kwargs = settings.model_dump()
        op = weave.op(fn, **op_kwargs)
        return op

    return wrapper


def get_mistral_patcher(
    settings: IntegrationSettings | None = None,
) -> MultiPatcher | NoOpPatcher:
    if settings is None:
        settings = IntegrationSettings()

    if not settings.enabled:
        return NoOpPatcher()

    global _mistral_patcher
    if _mistral_patcher is not None:
        return _mistral_patcher

    base = settings.op_settings

    chat_settings = base.model_copy(update={"name": base.name or "mistralai.chat"})
    chat_stream_settings = base.model_copy(
        update={"name": base.name or "mistralai.chat_stream"}
    )
    async_chat_settings = base.model_copy(
        update={"name": base.name or "mistralai.async_client.chat"}
    )
    async_chat_stream_settings = base.model_copy(
        update={"name": base.name or "mistralai.async_client.chat_stream"}
    )

    _mistral_patcher = MultiPatcher(
        [
            # Patch the sync, non-streaming chat method
            SymbolPatcher(
                lambda: importlib.import_module("mistralai.client"),
                "MistralClient.chat",
                mistral_wrapper(chat_settings),
            ),
            # Patch the sync, streaming chat method
            SymbolPatcher(
                lambda: importlib.import_module("mistralai.client"),
                "MistralClient.chat_stream",
                mistral_stream_wrapper(chat_stream_settings),
            ),
            # Patch the async, non-streaming chat method
            SymbolPatcher(
                lambda: importlib.import_module("mistralai.async_client"),
                "MistralAsyncClient.chat",
                mistral_wrapper(async_chat_settings),
            ),
            # Patch the async, streaming chat method
            SymbolPatcher(
                lambda: importlib.import_module("mistralai.async_client"),
                "MistralAsyncClient.chat_stream",
                mistral_stream_wrapper(async_chat_stream_settings),
            ),
        ]
    )

    return _mistral_patcher



import dataclasses
import inspect
import logging
import operator
import typing
from collections.abc import Generator, Iterator, Sequence
from copy import deepcopy
from typing import Any, Literal, Optional, SupportsIndex, Union

from pydantic import BaseModel
from pydantic import v1 as pydantic_v1

from weave.trace import box
from weave.trace.context.tests_context import get_raise_on_captured_errors
from weave.trace.context.weave_client_context import (
    get_weave_client,
    require_weave_client,
)
from weave.trace.object_record import ObjectRecord
from weave.trace.op import is_op, maybe_bind_method
from weave.trace.refs import (
    DICT_KEY_EDGE_NAME,
    LIST_INDEX_EDGE_NAME,
    OBJECT_ATTR_EDGE_NAME,
    TABLE_ROW_ID_EDGE_NAME,
    DeletedRef,
    ObjectRef,
    RefWithExtra,
    TableRef,
)
from weave.trace.serialization.serialize import from_json
from weave.trace.table import Table
from weave.trace_server.errors import ObjectDeletedError
from weave.trace_server.trace_server_interface import (
    ObjReadReq,
    TableQueryReq,
    TableQueryStatsReq,
    TableRowFilter,
    TraceServerInterface,
)
from weave.utils.iterators import ThreadSafeLazyList

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class MutationSetitem:
    path: tuple[str, ...]
    operation: Literal["setitem"]
    args: tuple[str, Any]


@dataclasses.dataclass
class MutationSetattr:
    path: tuple[str, ...]
    operation: Literal["setattr"]
    args: tuple[str, Any]


@dataclasses.dataclass
class MutationAppend:
    path: tuple[str, ...]
    operation: Literal["append"]
    args: tuple[Any]


Mutation = Union[MutationSetattr, MutationSetitem, MutationAppend]
MutationOperation = Union[Literal["setitem"], Literal["setattr"], Literal["append"]]


def make_mutation(
    path: tuple[str, ...], operation: MutationOperation, args: tuple[Any, ...]
) -> Mutation:
    if operation == "setitem":
        if len(args) != 2 or not isinstance(args[0], str):
            raise ValueError("setitem mutation requires 2 args")
        args = typing.cast(tuple[str, Any], args)
        return MutationSetitem(path, operation, args)
    elif operation == "setattr":
        if len(args) != 2 or not isinstance(args[0], str):
            raise ValueError("setattr mutation requires 2 args")
        args = typing.cast(tuple[str, Any], args)
        return MutationSetattr(path, operation, args)
    elif operation == "append":
        if len(args) != 1:
            raise ValueError("append mutation requires 1 arg")
        args = typing.cast(tuple[Any], args)
        return MutationAppend(path, operation, args)
    else:
        raise ValueError(f"Unknown operation: {operation}")


class Traceable:
    ref: Optional[RefWithExtra]
    mutations: Optional[list[Mutation]] = None
    root: "Traceable"
    parent: Optional["Traceable"] = None
    server: TraceServerInterface
    _is_dirty: bool = False

    def _mark_dirty(self) -> None:
        """Recursively mark this object and its ancestors as dirty and removes their refs."""
        self._is_dirty = True
        self.ref = None
        if (
            # Written this way to satisfy mypy
            self.parent is not self
            and self.parent is not None
            and hasattr(self.parent, "_mark_dirty")
        ):
            self.parent._mark_dirty()

    def add_mutation(
        self, path: tuple[str, ...], operation: MutationOperation, *args: Any
    ) -> None:
        if self.mutations is None:
            self.mutations = []
        self.mutations.append(make_mutation(path, operation, args))

    def save(self) -> ObjectRef:
        if not isinstance(self.ref, ObjectRef):
            raise TypeError("Can only save from object refs")
        if self.root is not self:
            raise ValueError("Can only save from root object")
        if self.mutations is None:
            raise ValueError("No mutations to save")

        mutations = self.mutations
        self.mutations = None
        raise NotImplementedError("Traceable.save not implemented")
        # return self.server.mutate(self.ref, mutations)


def pydantic_getattribute(self: BaseModel, name: str) -> Any:
    attribute = object.__getattribute__(self, name)

    # Starting in pydantic 2.10.0, this handling is needed otherwise getattribute will
    # infinitely recurse.
    if name.startswith("__") and name.endswith("__"):
        return attribute

    if name not in object.__getattribute__(self, "model_fields"):
        return attribute
    if name == "ref":
        try:
            return object.__getattribute__(self, "ref")
        except AttributeError:
            return None

    server = gc.server if (gc := get_weave_client()) else None
    res = attribute_access_result(self, attribute, name, server=server)

    # We need this because we override __getattribute__ and wrap the returned values.
    # The wrapped result may be mutable (e.g. list), so we need to replace the attribute
    # on self so that mutations are applied to the correct object and the user gets back
    # what they expect when they call `self.<name>`.
    self.__dict__[name] = res
    return res


def attribute_access_result(
    self: object,
    val_attr_val: Any,
    attr_name: str,
    *,
    server: Optional[TraceServerInterface],
) -> Any:
    # Not ideal, what about properties?
    if callable(val_attr_val):
        return maybe_bind_method(val_attr_val, self)

    if (ref := getattr(self, "ref", None)) is None:
        # Even if we have not parent ref, if our current value is an object
        # ref, we should still process it with make_trace_obj. Practically,
        # this allows a user to "get" a Model, update a field, then invoke it.
        # Primary test: `test_dirty_model_op_retrieval`
        if not isinstance(val_attr_val, ObjectRef):
            return val_attr_val
        new_ref = None
    else:
        new_ref = ref.with_attr(attr_name)

    if server is None:
        return val_attr_val

    root = getattr(self, "root", None)

    return make_trace_obj(
        val_attr_val,
        new_ref,
        server,
        root,
        self,
    )


class WeaveObject(Traceable):
    def __init__(
        self,
        val: Any,
        ref: Optional[RefWithExtra],
        server: TraceServerInterface,
        root: typing.Optional[Traceable],
        parent: Optional[Traceable] = None,
    ) -> None:
        self._val = val
        self.ref = ref
        self.server = server
        self.root = root or self
        self.parent = parent

    def __deepcopy__(self, memo: dict) -> "WeaveObject":
        val_copy = deepcopy(self._val, memo)
        res = WeaveObject(
            val_copy,
            ref=self.ref,  # maybe this should be zero'd?
            server=self.server,
            root=self.root,
        )
        memo[id(self)] = res
        return res

    def __getattribute__(self, __name: str) -> Any:
        try:
            return object.__getattribute__(self, __name)
        except AttributeError:
            pass
        val_attr_val = object.__getattribute__(self._val, __name)
        result = attribute_access_result(self, val_attr_val, __name, server=self.server)

        # Store the result on _val so we don't deref next time.
        try:
            object.__setattr__(self._val, __name, result)
        except AttributeError:
            # Happens if self._val.<name> is a property. Return the raw value instead
            # of a Traceable value.
            return val_attr_val
        return result

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name in [
            "_val",
            "ref",
            "server",
            "root",
            "mutations",
            "_is_dirty",
            "parent",
        ]:
            return object.__setattr__(self, __name, __value)
        else:
            self._mark_dirty()
            if isinstance(__value, Traceable):
                __value.parent = self

            return setattr(self._val, __name, __value)

    def __dir__(self) -> list[str]:
        return dir(self._val)

    def __repr__(self) -> str:
        return f"WeaveObject({self._val})"

    def __eq__(self, other: Any) -> bool:
        return self._val == other


class WeaveTable(Traceable):
    filter: TableRowFilter
    _known_length: Optional[int] = None
    _rows: Optional[Sequence[dict]] = None
    # _prefetched_rows is a local cache of rows that can be used to
    # avoid a remote call. Should only be used by internal code.
    _prefetched_rows: Optional[list[dict]] = None

    def __init__(
        self,
        table_ref: Optional[TableRef],
        ref: Optional[RefWithExtra],
        server: TraceServerInterface,
        filter: TableRowFilter,
        root: Optional[Traceable],
        parent: Optional[Traceable] = None,
    ) -> None:
        self.table_ref = table_ref
        self.filter = filter
        self.ref = ref  # type: ignore
        self.server = server
        self.root = root or self
        self.parent = parent

    @property
    def rows(self) -> Sequence[dict]:
        if self._rows is None:
            should_local_iter = (
                self.ref is not None
                and self.table_ref is not None
                and self.table_ref._row_digests is not None
                and self._prefetched_rows is not None
            )
            if should_local_iter:
                self._rows = ThreadSafeLazyList(self._local_iter_with_remote_fallback())
            else:
                self._rows = ThreadSafeLazyList(self._remote_iter())
        return self._rows

    @rows.setter
    def rows(self, value: Sequence[dict]) -> None:
        if not all(isinstance(row, dict) for row in value):
            raise ValueError("All table rows must be dicts")

        self._rows = value
        self._mark_dirty()

    def _inefficiently_materialize_rows_as_list(self) -> list[dict]:
        # This method is named `inefficiently` to warn callers that
        # it should be avoided. We have this nasty paradigm where sometimes
        # a WeaveTable needs to act like a list, but it is actually a remote
        # table. This method will force iteration through the remote data
        # and materialize it into a list. Any uses of this are signs of a design
        # problem arising from a remote table clashing with the need to feel like
        # a local list.
        if not isinstance(self.rows, list):
            self._rows = list(iter(self.rows))
            self._known_length = len(self._rows)
        return typing.cast(list[dict], self.rows)

    def set_prefetched_rows(self, prefetched_rows: list[dict]) -> None:
        """Sets the rows to a local cache of rows that can be used to
        avoid a remote call. Should only be used by internal code.

        It is expected that these rows are the exact same rows that would
        be returned by a query for this table. Failing to meet this expectation
        will cause table operations to behave unexpectedly.
        """
        if self._rows is not None:
            raise ValueError(
                "Cannot set prefetched rows on WeaveTable when rows are already loaded"
            )
        self._prefetched_rows = prefetched_rows

    def __len__(self) -> int:
        # This should be a single query
        if self._known_length is not None:
            return self._known_length

        # Condition 1: we already have all the rows in memory
        if self._prefetched_rows is not None:
            self._known_length = len(self._prefetched_rows)
            return self._known_length

        # Condition 2: we have the row digests and they are a list
        if (
            self.table_ref is not None
            and self.table_ref._row_digests is not None
            and isinstance(self.table_ref._row_digests, list)
        ):
            self._known_length = len(self.table_ref._row_digests)
            return self._known_length

        # Condition 3: We don't know the length, in which case we can get it from the server
        if self.table_ref is not None:
            self._known_length = self._fetch_remote_length()
            return self._known_length

        # Finally, if we have no table ref, we can still get the length
        # by materializing the rows as a list. I actually think this
        # can never happen, but it is here for completeness.
        rows_as_list = self._inefficiently_materialize_rows_as_list()
        return len(rows_as_list)

    def _fetch_remote_length(self) -> int:
        if self.table_ref is None:
            raise ValueError("Cannot fetch remote length of table without table ref")

        response = self.server.table_query_stats(
            TableQueryStatsReq(
                project_id=self.table_ref.project_id, digest=self.table_ref.digest
            )
        )
        return response.count

    def __eq__(self, other: Any) -> bool:
        rows = self._inefficiently_materialize_rows_as_list()
        return rows == other

    def _mark_dirty(self) -> None:
        self.table_ref = None
        self._prefetched_rows = None
        self._known_length = None
        super()._mark_dirty()

    def _local_iter_with_remote_fallback(self) -> Generator[dict, None, None]:
        """
        This is the case where we:
        1. Have all the rows in memory
        2. Have all the row digests

        In this case, we don't need to make any calls and can just return the rows
        """
        wc = get_weave_client()
        if (
            wc is None
            or self.ref is None
            or self.table_ref is None
            or self.table_ref._row_digests is None
            or self._prefetched_rows is None
        ):
            if get_raise_on_captured_errors():
                raise
            logger.error(
                "Expected all row digests and prefetched rows to be set, falling back to remote iteration"
            )
            yield from self._remote_iter()
            return

        cached_table_ref = self.table_ref
        if isinstance(self.table_ref._row_digests, list):
            # Only do this check if it is resolved
            row_digest_len = len(self.table_ref._row_digests)
            prefetched_rows_len = len(self._prefetched_rows)
            if row_digest_len != prefetched_rows_len:
                if get_raise_on_captured_errors():
                    raise
                logger.error(
                    f"Expected length of row digests ({row_digest_len}) to match prefetched rows ({prefetched_rows_len}). Falling back to remote iteration."
                )
                yield from self._remote_iter()
                return

        for i, _ in enumerate(self._prefetched_rows):
            next_id_future = wc.future_executor.defer(
                lambda closure=i: cached_table_ref.row_digests[closure]
            )
            new_ref = self.ref.with_item(next_id_future)
            val = self._prefetched_rows[i]
            res = from_json(val, self.table_ref.project_id, self.server)
            res = make_trace_obj(res, new_ref, self.server, self.root)
            yield res

    def _remote_iter(self) -> Generator[dict, None, None]:
        if self.table_ref is None:
            return

        wc = require_weave_client()

        page_index = 0
        page_size = 100
        while True:
            response = self.server.table_query(
                TableQueryReq(
                    project_id=self.table_ref.project_id,
                    digest=self.table_ref.digest,
                    offset=page_index * page_size,
                    limit=page_size,
                    filter=self.filter,
                )
            )

            # When paginating through large datasets, we need special handling for prefetched rows
            # on the first page. This is because prefetched_rows contains ALL rows, while each
            # response page contains at most page_size rows.
            if page_index == 0 and self._prefetched_rows is not None:
                response_rows_len = len(response.rows)
                prefetched_rows_len = len(self._prefetched_rows)

                # There are two valid scenarios:
                # 1. The response rows exactly match prefetched rows (small dataset, no pagination needed)
                # 2. We're paginating a large dataset (response has page_size rows, prefetched has more)
                #
                # Any other mismatch indicates an inconsistency that should be handled by
                # discarding the prefetched rows and relying solely on server responses.
                if response_rows_len != prefetched_rows_len and not (
                    response_rows_len == page_size and prefetched_rows_len > page_size
                ):
                    msg = f"Expected length of response rows ({response_rows_len}) to match prefetched rows ({prefetched_rows_len}). Ignoring prefetched rows."
                    if get_raise_on_captured_errors():
                        raise ValueError(msg)
                    logger.debug(msg)
                    self._prefetched_rows = None

            # Process rows in parallel using the weave client's future executor
            futures = []
            for i, item in enumerate(response.rows):
                new_ref = self.ref.with_item(item.digest) if self.ref else None
                # Here, we use the raw rows if they exist, otherwise we use the
                # rows from the server. This is a temporary trick to ensure
                # we don't re-deserialize the rows on every access. Once all servers
                # return digests, this branch can be removed because anytime we have prefetched
                # rows we should also have the digests - and we should be in the
                #  _local_iter_with_remote_fallback case.
                val = (
                    item.val
                    if self._prefetched_rows is None
                    else self._prefetched_rows[page_index * page_size + i]
                )

                def process_row(val: Any, new_ref: RefWithExtra) -> Any:
                    if not self.table_ref:
                        # Should never happen, we need table_ref to remote_iter
                        return None
                    return make_trace_obj(
                        from_json(val, self.table_ref.project_id, self.server),
                        new_ref,
                        self.server,
                        self.root,
                    )

                future = wc.future_executor.defer(
                    lambda v=val, r=new_ref: process_row(v, r)
                )
                futures.append(future)

            # Yield results as they complete
            for future in futures:
                yield future.result()

            if len(response.rows) < page_size:
                break

            page_index += 1

    def __getitem__(self, key: Union[int, slice, str]) -> Any:
        # TODO: ideally we would have some sort of intelligent
        # LRU style caching that allows us to minimize materialization
        # of the rows as a list.
        rows = self._inefficiently_materialize_rows_as_list()

        if isinstance(key, (int, slice)):
            return rows[key]

        for row in rows:
            if row.ref.extra[-1] == key:  # type: ignore
                return row

        raise KeyError(f"Row ID not found: {key}")

    def __iter__(self) -> Iterator[dict]:
        return iter(self.rows)

    def append(self, val: dict) -> None:
        rows = self._inefficiently_materialize_rows_as_list()
        if not isinstance(val, dict):
            raise TypeError("Can only append dicts to tables")
        self._mark_dirty()
        rows.append(val)

    def pop(self, index: int) -> None:
        rows = self._inefficiently_materialize_rows_as_list()
        self._mark_dirty()
        rows.pop(index)


class WeaveList(Traceable, list):
    def __init__(
        self,
        *args: Any,
        server: TraceServerInterface,
        ref: Optional[RefWithExtra] = None,
        root: Optional[Traceable] = None,
        parent: Optional[Traceable] = None,
    ) -> None:
        self.server = server

        self.ref = ref
        self.root = root if root is not None else self
        self.parent = parent
        super().__init__(*args)

    def __deepcopy__(self, memo: dict) -> "WeaveList":
        items_copy = [deepcopy(item, memo) for item in self]
        res = WeaveList(
            items_copy,
            server=self.server,
            ref=self.ref,  # maybe this should be zero'd?
            root=self.root,
            parent=self.parent,
        )
        memo[id(self)] = res
        return res

    def __getitem__(self, i: Union[SupportsIndex, slice]) -> Any:
        if isinstance(i, slice):
            raise TypeError("Slices not yet supported")
        index = operator.index(i)
        new_ref = self.ref.with_index(index) if self.ref else None
        index_val = super().__getitem__(index)
        return make_trace_obj(index_val, new_ref, self.server, self.root)

    def __setitem__(self, i: Union[SupportsIndex, slice], value: Any) -> None:
        if isinstance(i, slice):
            raise TypeError("Slices not yet supported")
        if (index := operator.index(i)) >= len(self):
            raise IndexError("list assignment index out of range")

        # Though this ostensibly only marks the parent (list) as dirty, siblings
        # will also get new refs because their old refs are relative to the parent
        # (the element refs will be extras of the new parent ref)
        self._mark_dirty()
        if isinstance(value, Traceable):
            value.parent = self

        super().__setitem__(index, value)

    def append(self, item: Any) -> None:
        self._mark_dirty()
        if isinstance(item, Traceable):
            item.parent = self

        super().append(item)

    def __iter__(self) -> Iterator[Any]:
        for i in range(len(self)):
            yield self[i]

    def __repr__(self) -> str:
        return f"WeaveList({super().__repr__()})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, list):
            return False
        if len(self) != len(other):
            return False
        for v1, v2 in zip(self, other):
            if v1 != v2:
                return False
        return True


class WeaveDict(Traceable, dict):
    def __init__(
        self,
        *args: Any,
        server: TraceServerInterface,
        ref: Optional[RefWithExtra] = None,
        root: Optional[Traceable] = None,
        parent: Optional[Traceable] = None,
        **kwargs: Any,
    ) -> None:
        self.server = server

        self.ref = ref
        self.root = root if root is not None else self
        self.parent = parent
        super().__init__(*args, **kwargs)

    def __deepcopy__(self, memo: dict) -> "WeaveDict":
        items_copy = {k: deepcopy(v, memo) for k, v in self.items()}
        res = WeaveDict(
            items_copy,
            server=self.server,
            ref=self.ref,  # maybe this should be zero'd?
            root=self.root,
            parent=self.parent,
        )
        memo[id(self)] = res
        return res

    def __getitem__(self, key: str) -> Any:
        new_ref = self.ref.with_key(key) if self.ref else None
        v = super().__getitem__(key)
        return make_trace_obj(v, new_ref, self.server, self.root)

    def get(self, key: str, default: Any = None) -> Any:
        new_ref = self.ref.with_key(key) if self.ref else None
        v = super().get(key, default)
        return make_trace_obj(v, new_ref, self.server, self.root)

    def __setitem__(self, key: str, value: Any) -> None:
        # Though this ostensibly only marks the parent (dict) as dirty, siblings
        # will also get new refs because their old refs are relative to the parent
        # (the element refs will be extras of the new parent ref)
        self._mark_dirty()
        if isinstance(value, Traceable):
            value.parent = self

        super().__setitem__(key, value)

    def keys(self) -> Generator[Any, Any, Any]:  # type: ignore
        yield from super().keys()

    def values(self) -> Generator[Any, Any, Any]:  # type: ignore
        for k in self.keys():
            yield self[k]

    def items(self) -> Generator[tuple[Any, Any], Any, Any]:  # type: ignore
        for k in self.keys():
            yield k, self[k]

    def __iter__(self) -> Iterator[str]:
        # Simply define this to so that d = WeaveDict({'a': 1, 'b': 2})); d2 = dict(d)
        # works. The dict(d) constructor works differently if __iter__ is not defined
        # on d.
        return super().__iter__()

    def __repr__(self) -> str:
        return f"WeaveDict({super().__repr__()})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, dict):
            return False
        if len(self) != len(other):
            return False
        for k, v in self.items():
            if k not in other:
                return False
            if other[k] != v:
                return False
        return True


class InternalError(Exception): ...


def make_trace_obj(
    val: Any,
    new_ref: Optional[RefWithExtra],  # Can this actually be None?
    server: TraceServerInterface,
    root: Optional[Traceable],
    parent: Any = None,
) -> Any:
    if isinstance(val, Traceable):
        # If val is a WeaveTable, we want to refer to it via the outer object
        # that it is within, rather than via the TableRef. For example we
        # want Dataset row refs to be Dataset.rows[id] rather than table[id]
        if isinstance(val, WeaveTable):
            val.ref = new_ref
        return val
    if hasattr(val, "ref") and isinstance(val.ref, RefWithExtra):
        # The Traceable check above does not currently work for Ops, where we
        # directly attach a ref, or to our Boxed classes. We should use Traceable
        # for all of these, but for now we need to check for the ref attribute.
        return val
    # Dereference val and create the appropriate wrapper object
    extra: tuple[str, ...] = ()
    if isinstance(val, ObjectRef):
        new_ref = val
        extra = val.extra
        try:
            project_id = f"{val.entity}/{val.project}"
            read_res = server.obj_read(
                ObjReadReq(
                    project_id=project_id,
                    object_id=val.name,
                    digest=val.digest,
                )
            )
            val = from_json(read_res.obj.val, project_id, server)
        except ObjectDeletedError as e:
            # encountered a deleted object, return DeletedRef, warn and continue
            val = DeletedRef(ref=new_ref, deleted_at=e.deleted_at, error=e)
            logger.warning(f"Could not read deleted object: {new_ref}")

    if isinstance(val, Table):
        val_ref = val.ref
        if not isinstance(val_ref, TableRef):
            val_table_ref = getattr(val, "table_ref", None)
            if not isinstance(val_table_ref, TableRef):
                raise InternalError(
                    "Expected Table.ref or Table.table_ref to be TableRef"
                )
            val_ref = val_table_ref
        rows = val.rows
        val = WeaveTable(
            table_ref=val_ref,
            ref=new_ref,
            server=server,
            filter=TableRowFilter(),
            root=root,
            parent=parent,
        )
        # Use in memory rows! This is the case where we are making
        # a trace object from an existing Table! If we don't do this
        # then the WeaveTable will try to fetch all the rows from the
        # server, throwing away the in memory rows. This is really expensive
        # when we are doing evaluations!
        val.set_prefetched_rows(rows)
    if isinstance(val, TableRef):
        val = WeaveTable(
            table_ref=val,
            ref=new_ref,
            server=server,
            filter=TableRowFilter(),
            root=root,
            parent=parent,
        )

    if extra:
        # This is where extra resolution happens?
        for extra_index in range(0, len(extra), 2):
            op, arg = extra[extra_index], extra[extra_index + 1]
            if op == DICT_KEY_EDGE_NAME:
                val = val[arg]
            elif op == OBJECT_ATTR_EDGE_NAME:
                val = getattr(val, arg)
            elif op == LIST_INDEX_EDGE_NAME:
                val = val[int(arg)]
            elif op == TABLE_ROW_ID_EDGE_NAME:
                val = val[arg]
            else:
                raise ValueError(f"Unknown ref type: {extra[extra_index]}")

            # need to deref if we encounter these
            if isinstance(val, TableRef):
                table_row_filter = TableRowFilter()
                if (
                    len(extra) == 4
                    and extra[0] == OBJECT_ATTR_EDGE_NAME
                    and extra[1] == "rows"
                    and extra[2] == TABLE_ROW_ID_EDGE_NAME
                ):
                    table_row_filter.row_digests = [extra[3]]

                val = WeaveTable(
                    table_ref=val,
                    ref=new_ref,
                    server=server,
                    filter=table_row_filter,
                    root=root,
                    parent=parent,
                )

    if not isinstance(val, Traceable):
        if isinstance(val, ObjectRecord):
            return WeaveObject(
                val, ref=new_ref, server=server, root=root, parent=parent
            )
        elif isinstance(val, list):
            return WeaveList(val, ref=new_ref, server=server, root=root, parent=parent)
        elif isinstance(val, dict):
            return WeaveDict(val, ref=new_ref, server=server, root=root)
    if is_op(val) and inspect.signature(val.resolve_fn).parameters.get("self"):
        # This condition attempts to bind the current `self` to the attribute if
        # it happens to be both an `Op` and have a `self` argument. This is a
        # bit of a hack since we are not always sure that the current object is
        # the correct object to bind. There are 3 cases:
        # 1. The attribute is part of the instance methods and the binding is
        #    correct
        # 2. The attribute is assigned as a property and is not bound at
        #    assignment time. In this case, it is "unlikely" that the args
        #    contain a `self` argument - which is why we apply this heuristic.
        # 3. The attribute is assigned as a property and is bound to another
        #    object at the time of assignment. In this case, the binding is
        #    incorrect. However, in our evaluation use case we do not have this
        #    case. We are accepting the incorrect assignment here for the sake
        #    of expediency, but should be fixed.
        if parent is None:
            raise MissingSelfInstanceError(
                f"{val.name} Op requires a bound self instance. Must be called from an instance method."
            )
        # TODO: This binding is correct but not done for consistency with the
        # not-yet-saved-method-op API which requires explicitly passing self
        # val.call = partial(call, val, parent)
        val = maybe_bind_method(val, parent)
    box_val = box.box(val)
    if isinstance(box_val, pydantic_v1.BaseModel) or is_op(val):
        box_val.__dict__["ref"] = new_ref
    elif box_val is None or isinstance(box_val, bool):
        # We intentionally don't box None and bools because it's imposible to
        # make them behave like the underlying True/False/None objects in python.
        # This is unlike other objects (dict, list, int) that can be inherited
        # from and compared.

        # The tradeoff we're making here is:
        # 1. We won't ref track bools or None when passed into a call; but
        # 2. Users can compare them pythonically (e.g. `x is None` vs. `x == None`)

        pass
    else:
        if hasattr(box_val, "ref") and not isinstance(box_val, DeletedRef):
            setattr(box_val, "ref", new_ref)
    return box_val


class MissingSelfInstanceError(ValueError):
    pass



import io
import json
import logging
from collections.abc import Iterator
from typing import Any, Optional, Union, cast

from pydantic import BaseModel

from weave.trace.env import weave_trace_server_url
from weave.trace.settings import max_calls_queue_size
from weave.trace_server import requests
from weave.trace_server import trace_server_interface as tsi
from weave.trace_server_bindings.async_batch_processor import AsyncBatchProcessor
from weave.utils.retry import with_retry
from weave.wandb_interface import project_creator

logger = logging.getLogger(__name__)

# Default timeout values (in seconds)
# DEFAULT_CONNECT_TIMEOUT = 10
# DEFAULT_READ_TIMEOUT = 30
# DEFAULT_TIMEOUT = (DEFAULT_CONNECT_TIMEOUT, DEFAULT_READ_TIMEOUT)


class StartBatchItem(BaseModel):
    mode: str = "start"
    req: tsi.CallStartReq


class EndBatchItem(BaseModel):
    mode: str = "end"
    req: tsi.CallEndReq


class Batch(BaseModel):
    batch: list[Union[StartBatchItem, EndBatchItem]]


class ServerInfoRes(BaseModel):
    min_required_weave_python_version: str


REMOTE_REQUEST_BYTES_LIMIT = (
    (32 - 1) * 1024 * 1024
)  # 32 MiB (real limit) - 1 MiB (buffer)


class RemoteHTTPTraceServer(tsi.TraceServerInterface):
    trace_server_url: str

    # My current batching is not safe in notebooks, disable it for now
    def __init__(
        self,
        trace_server_url: str,
        should_batch: bool = False,
        *,
        remote_request_bytes_limit: int = REMOTE_REQUEST_BYTES_LIMIT,
    ):
        super().__init__()
        self.trace_server_url = trace_server_url
        self.should_batch = should_batch
        self.call_processor = None
        if self.should_batch:
            self.call_processor = AsyncBatchProcessor(
                self._flush_calls,
                max_queue_size=max_calls_queue_size(),
            )
        self._auth: Optional[tuple[str, str]] = None
        self.remote_request_bytes_limit = remote_request_bytes_limit

    def ensure_project_exists(
        self, entity: str, project: str
    ) -> tsi.EnsureProjectExistsRes:
        # TODO: This should happen in the wandb backend, not here, and it's slow
        # (hundreds of ms)
        return tsi.EnsureProjectExistsRes.model_validate(
            project_creator.ensure_project_exists(entity, project)
        )

    @classmethod
    def from_env(cls, should_batch: bool = False) -> "RemoteHTTPTraceServer":
        # Explicitly calling `RemoteHTTPTraceServer` constructor here to ensure
        # that type checking is applied to the constructor.
        return RemoteHTTPTraceServer(weave_trace_server_url(), should_batch)

    def set_auth(self, auth: tuple[str, str]) -> None:
        self._auth = auth

    @with_retry
    def _send_batch_to_server(self, encoded_data: bytes) -> None:
        """Send a batch of data to the server with retry logic.

        This method is separated from _flush_calls to avoid recursive retries.
        """
        r = requests.post(
            self.trace_server_url + "/call/upsert_batch",
            data=encoded_data,  # type: ignore
            auth=self._auth,
            # timeout=DEFAULT_TIMEOUT,
        )
        if r.status_code == 413:
            # handle 413 explicitly to provide actionable error message
            reason = json.loads(r.text)["reason"]
            raise requests.HTTPError(f"413 Client Error: {reason}", response=r)
        r.raise_for_status()

    def _flush_calls(
        self,
        batch: list[Union[StartBatchItem, EndBatchItem]],
        *,
        _should_update_batch_size: bool = True,
    ) -> None:
        """Process a batch of calls, splitting if necessary and sending to the server.

        This method handles the logic of splitting batches that are too large,
        but delegates the actual server communication (with retries) to _send_batch_to_server.
        """
        # Call processor must be defined for this method
        assert self.call_processor is not None
        if len(batch) == 0:
            return

        data = Batch(batch=batch).model_dump_json()
        encoded_data = data.encode("utf-8")
        encoded_bytes = len(encoded_data)

        # Update target batch size (this allows us to have a dynamic batch size based on the size of the data being sent)
        estimated_bytes_per_item = encoded_bytes / len(batch)
        if _should_update_batch_size and estimated_bytes_per_item > 0:
            target_batch_size = int(
                self.remote_request_bytes_limit // estimated_bytes_per_item
            )
            self.call_processor.max_batch_size = max(1, target_batch_size)

        # If the batch is too big, split it in half and process each half
        if encoded_bytes > self.remote_request_bytes_limit and len(batch) > 1:
            split_idx = int(len(batch) // 2)
            self._flush_calls(batch[:split_idx], _should_update_batch_size=False)
            self._flush_calls(batch[split_idx:], _should_update_batch_size=False)
            return

        # If a single item is too large, we can't send it -- log an error and drop it
        if encoded_bytes > self.remote_request_bytes_limit and len(batch) == 1:
            logger.error(
                f"Single call size ({encoded_bytes} bytes) is too large to send. "
                f"The maximum size is {self.remote_request_bytes_limit} bytes."
            )

        try:
            self._send_batch_to_server(encoded_data)
        except Exception:
            # Add items back to the queue for later processing
            logger.warning(
                f"Batch failed after max retries, requeueing batch with {len(batch)=} for later processing",
            )

            # only if debug mode
            if logger.isEnabledFor(logging.DEBUG):
                ids = []
                for item in batch:
                    if isinstance(item, StartBatchItem):
                        ids.append(f"{item.req.start.id}-start")
                    elif isinstance(item, EndBatchItem):
                        ids.append(f"{item.req.end.id}-end")
                logger.debug(f"Requeueing batch with {ids=}")
            self.call_processor.enqueue(batch)

    @with_retry
    def _generic_request_executor(
        self,
        url: str,
        req: BaseModel,
        stream: bool = False,
    ) -> requests.Response:
        r = requests.post(
            self.trace_server_url + url,
            # `by_alias` is required since we have Mongo-style properties in the
            # query models that are aliased to conform to start with `$`. Without
            # this, the model_dump will use the internal property names which are
            # not valid for the `model_validate` step.
            data=req.model_dump_json(by_alias=True).encode("utf-8"),
            auth=self._auth,
            stream=stream,
            # timeout=DEFAULT_TIMEOUT,
        )
        if r.status_code == 500:
            reason_val = r.text
            try:
                reason_val = json.dumps(json.loads(reason_val), indent=2)
            except json.JSONDecodeError:
                reason_val = f"Reason: {reason_val}"
            raise requests.HTTPError(
                f"500 Server Error: Internal Server Error for url: {url}. {reason_val}",
                response=r,
            )
        r.raise_for_status()

        return r

    def _generic_request(
        self,
        url: str,
        req: BaseModel,
        req_model: type[BaseModel],
        res_model: type[BaseModel],
    ) -> BaseModel:
        if isinstance(req, dict):
            req = req_model.model_validate(req)
        r = self._generic_request_executor(url, req)
        return res_model.model_validate(r.json())

    def _generic_stream_request(
        self,
        url: str,
        req: BaseModel,
        req_model: type[BaseModel],
        res_model: type[BaseModel],
    ) -> Iterator[BaseModel]:
        if isinstance(req, dict):
            req = req_model.model_validate(req)
        r = self._generic_request_executor(url, req, stream=True)
        for line in r.iter_lines():
            if line:
                yield res_model.model_validate_json(line)

    @with_retry
    def server_info(self) -> ServerInfoRes:
        r = requests.get(
            self.trace_server_url + "/server_info",
            # timeout=DEFAULT_TIMEOUT,
        )
        r.raise_for_status()
        return ServerInfoRes.model_validate(r.json())

    # Call API
    def call_start(
        self, req: Union[tsi.CallStartReq, dict[str, Any]]
    ) -> tsi.CallStartRes:
        if self.should_batch:
            assert self.call_processor is not None

            req_as_obj: tsi.CallStartReq
            if isinstance(req, dict):
                req_as_obj = tsi.CallStartReq.model_validate(req)
            else:
                req_as_obj = req
            if req_as_obj.start.id == None or req_as_obj.start.trace_id == None:
                raise ValueError(
                    "CallStartReq must have id and trace_id when batching."
                )
            self.call_processor.enqueue([StartBatchItem(req=req_as_obj)])
            return tsi.CallStartRes(
                id=req_as_obj.start.id, trace_id=req_as_obj.start.trace_id
            )
        return self._generic_request(
            "/call/start", req, tsi.CallStartReq, tsi.CallStartRes
        )

    def call_start_batch(self, req: tsi.CallCreateBatchReq) -> tsi.CallCreateBatchRes:
        return self._generic_request(
            "/call/upsert_batch", req, tsi.CallCreateBatchReq, tsi.CallCreateBatchRes
        )

    def call_end(self, req: Union[tsi.CallEndReq, dict[str, Any]]) -> tsi.CallEndRes:
        if self.should_batch:
            assert self.call_processor is not None

            req_as_obj: tsi.CallEndReq
            if isinstance(req, dict):
                req_as_obj = tsi.CallEndReq.model_validate(req)
            else:
                req_as_obj = req
            self.call_processor.enqueue([EndBatchItem(req=req_as_obj)])
            return tsi.CallEndRes()
        return self._generic_request("/call/end", req, tsi.CallEndReq, tsi.CallEndRes)

    def call_read(self, req: Union[tsi.CallReadReq, dict[str, Any]]) -> tsi.CallReadRes:
        return self._generic_request(
            "/call/read", req, tsi.CallReadReq, tsi.CallReadRes
        )

    def calls_query(
        self, req: Union[tsi.CallsQueryReq, dict[str, Any]]
    ) -> tsi.CallsQueryRes:
        # This previously called the deprecated /calls/query endpoint.
        return tsi.CallsQueryRes(calls=list(self.calls_query_stream(req)))

    def calls_query_stream(
        self, req: Union[tsi.CallsQueryReq, dict[str, Any]]
    ) -> Iterator[tsi.CallSchema]:
        return self._generic_stream_request(
            "/calls/stream_query", req, tsi.CallsQueryReq, tsi.CallSchema
        )

    def calls_query_stats(
        self, req: Union[tsi.CallsQueryStatsReq, dict[str, Any]]
    ) -> tsi.CallsQueryStatsRes:
        return self._generic_request(
            "/calls/query_stats", req, tsi.CallsQueryStatsReq, tsi.CallsQueryStatsRes
        )

    def calls_delete(
        self, req: Union[tsi.CallsDeleteReq, dict[str, Any]]
    ) -> tsi.CallsDeleteRes:
        return self._generic_request(
            "/calls/delete", req, tsi.CallsDeleteReq, tsi.CallsDeleteRes
        )

    def call_update(
        self, req: Union[tsi.CallUpdateReq, dict[str, Any]]
    ) -> tsi.CallUpdateRes:
        return self._generic_request(
            "/call/update", req, tsi.CallUpdateReq, tsi.CallUpdateRes
        )

    # Op API

    def op_create(self, req: Union[tsi.OpCreateReq, dict[str, Any]]) -> tsi.OpCreateRes:
        return self._generic_request(
            "/op/create", req, tsi.OpCreateReq, tsi.OpCreateRes
        )

    def op_read(self, req: Union[tsi.OpReadReq, dict[str, Any]]) -> tsi.OpReadRes:
        return self._generic_request("/op/read", req, tsi.OpReadReq, tsi.OpReadRes)

    def ops_query(self, req: Union[tsi.OpQueryReq, dict[str, Any]]) -> tsi.OpQueryRes:
        return self._generic_request("/ops/query", req, tsi.OpQueryReq, tsi.OpQueryRes)

    # Obj API

    def obj_create(
        self, req: Union[tsi.ObjCreateReq, dict[str, Any]]
    ) -> tsi.ObjCreateRes:
        return self._generic_request(
            "/obj/create", req, tsi.ObjCreateReq, tsi.ObjCreateRes
        )

    def obj_read(self, req: Union[tsi.ObjReadReq, dict[str, Any]]) -> tsi.ObjReadRes:
        return self._generic_request("/obj/read", req, tsi.ObjReadReq, tsi.ObjReadRes)

    def objs_query(
        self, req: Union[tsi.ObjQueryReq, dict[str, Any]]
    ) -> tsi.ObjQueryRes:
        return self._generic_request(
            "/objs/query", req, tsi.ObjQueryReq, tsi.ObjQueryRes
        )

    def obj_delete(self, req: tsi.ObjDeleteReq) -> tsi.ObjDeleteRes:
        return self._generic_request(
            "/obj/delete", req, tsi.ObjDeleteReq, tsi.ObjDeleteRes
        )

    def table_create(
        self, req: Union[tsi.TableCreateReq, dict[str, Any]]
    ) -> tsi.TableCreateRes:
        """Similar to `calls/batch_upsert`, we can dynamically adjust the payload size
        due to the property that table creation can be decomposed into a series of
        updates. This is useful when the table creation size is too big to be sent in
        a single request. We can create an empty table first, then update the table
        with the rows.
        """
        if isinstance(req, dict):
            req = tsi.TableCreateReq.model_validate(req)
        req = cast(tsi.TableCreateReq, req)

        estimated_bytes = len(req.model_dump_json(by_alias=True).encode("utf-8"))
        if estimated_bytes > self.remote_request_bytes_limit:
            initialization_req = tsi.TableCreateReq(
                table=tsi.TableSchemaForInsert(
                    project_id=req.table.project_id,
                    rows=[],
                )
            )
            initialization_res = self.table_create(initialization_req)

            update_req = tsi.TableUpdateReq(
                project_id=req.table.project_id,
                base_digest=initialization_res.digest,
                updates=[
                    tsi.TableAppendSpec(append=tsi.TableAppendSpecPayload(row=row))
                    for row in req.table.rows
                ],
            )
            update_res = self.table_update(update_req)

            return tsi.TableCreateRes(
                digest=update_res.digest, row_digests=update_res.updated_row_digests
            )
        else:
            return self._generic_request(
                "/table/create", req, tsi.TableCreateReq, tsi.TableCreateRes
            )

    def table_update(self, req: tsi.TableUpdateReq) -> tsi.TableUpdateRes:
        """Similar to `calls/batch_upsert`, we can dynamically adjust the payload size
        due to the property that table updates can be decomposed into a series of
        updates.
        """
        if isinstance(req, dict):
            req = tsi.TableUpdateReq.model_validate(req)
        req = cast(tsi.TableUpdateReq, req)

        estimated_bytes = len(req.model_dump_json(by_alias=True).encode("utf-8"))
        if estimated_bytes > self.remote_request_bytes_limit and len(req.updates) > 1:
            split_ndx = len(req.updates) // 2
            first_half_req = tsi.TableUpdateReq(
                project_id=req.project_id,
                base_digest=req.base_digest,
                updates=req.updates[:split_ndx],
            )
            first_half_res = self.table_update(first_half_req)
            second_half_req = tsi.TableUpdateReq(
                project_id=req.project_id,
                base_digest=first_half_res.digest,
                updates=req.updates[split_ndx:],
            )
            second_half_res = self.table_update(second_half_req)
            all_digests = (
                first_half_res.updated_row_digests + second_half_res.updated_row_digests
            )
            return tsi.TableUpdateRes(
                digest=second_half_res.digest, updated_row_digests=all_digests
            )
        else:
            return self._generic_request(
                "/table/update", req, tsi.TableUpdateReq, tsi.TableUpdateRes
            )

    def table_query(
        self, req: Union[tsi.TableQueryReq, dict[str, Any]]
    ) -> tsi.TableQueryRes:
        return self._generic_request(
            "/table/query", req, tsi.TableQueryReq, tsi.TableQueryRes
        )

    def table_query_stream(
        self, req: tsi.TableQueryReq
    ) -> Iterator[tsi.TableRowSchema]:
        # Need to manually iterate over this until the stram endpoint is built and shipped.
        res = self.table_query(req)
        yield from res.rows

    def table_query_stats(
        self, req: Union[tsi.TableQueryStatsReq, dict[str, Any]]
    ) -> tsi.TableQueryStatsRes:
        return self._generic_request(
            "/table/query_stats", req, tsi.TableQueryStatsReq, tsi.TableQueryStatsRes
        )

    def refs_read_batch(
        self, req: Union[tsi.RefsReadBatchReq, dict[str, Any]]
    ) -> tsi.RefsReadBatchRes:
        return self._generic_request(
            "/refs/read_batch", req, tsi.RefsReadBatchReq, tsi.RefsReadBatchRes
        )

    @with_retry
    def file_create(self, req: tsi.FileCreateReq) -> tsi.FileCreateRes:
        r = requests.post(
            self.trace_server_url + "/files/create",
            auth=self._auth,
            data={"project_id": req.project_id},
            files={"file": (req.name, req.content)},
            # timeout=DEFAULT_TIMEOUT,
        )
        r.raise_for_status()
        return tsi.FileCreateRes.model_validate(r.json())

    @with_retry
    def file_content_read(self, req: tsi.FileContentReadReq) -> tsi.FileContentReadRes:
        r = requests.post(
            self.trace_server_url + "/files/content",
            json={"project_id": req.project_id, "digest": req.digest},
            auth=self._auth,
            # timeout=DEFAULT_TIMEOUT,
        )
        r.raise_for_status()
        # TODO: Should stream to disk rather than to memory
        bytes = io.BytesIO()
        bytes.writelines(r.iter_content())
        bytes.seek(0)
        return tsi.FileContentReadRes(content=bytes.read())

    def feedback_create(
        self, req: Union[tsi.FeedbackCreateReq, dict[str, Any]]
    ) -> tsi.FeedbackCreateRes:
        return self._generic_request(
            "/feedback/create", req, tsi.FeedbackCreateReq, tsi.FeedbackCreateRes
        )

    def feedback_query(
        self, req: Union[tsi.FeedbackQueryReq, dict[str, Any]]
    ) -> tsi.FeedbackQueryRes:
        return self._generic_request(
            "/feedback/query", req, tsi.FeedbackQueryReq, tsi.FeedbackQueryRes
        )

    def feedback_purge(
        self, req: Union[tsi.FeedbackPurgeReq, dict[str, Any]]
    ) -> tsi.FeedbackPurgeRes:
        return self._generic_request(
            "/feedback/purge", req, tsi.FeedbackPurgeReq, tsi.FeedbackPurgeRes
        )

    def feedback_replace(
        self, req: Union[tsi.FeedbackReplaceReq, dict[str, Any]]
    ) -> tsi.FeedbackReplaceRes:
        return self._generic_request(
            "/feedback/replace", req, tsi.FeedbackReplaceReq, tsi.FeedbackReplaceRes
        )

    def actions_execute_batch(
        self, req: Union[tsi.ActionsExecuteBatchReq, dict[str, Any]]
    ) -> tsi.ActionsExecuteBatchRes:
        return self._generic_request(
            "/actions/execute_batch",
            req,
            tsi.ActionsExecuteBatchReq,
            tsi.ActionsExecuteBatchRes,
        )

    # Cost API
    def cost_query(
        self, req: Union[tsi.CostQueryReq, dict[str, Any]]
    ) -> tsi.CostQueryRes:
        return self._generic_request(
            "/cost/query", req, tsi.CostQueryReq, tsi.CostQueryRes
        )

    def cost_create(
        self, req: Union[tsi.CostCreateReq, dict[str, Any]]
    ) -> tsi.CostCreateRes:
        return self._generic_request(
            "/cost/create", req, tsi.CostCreateReq, tsi.CostCreateRes
        )

    def cost_purge(
        self, req: Union[tsi.CostPurgeReq, dict[str, Any]]
    ) -> tsi.CostPurgeRes:
        return self._generic_request(
            "/cost/purge", req, tsi.CostPurgeReq, tsi.CostPurgeRes
        )

    def completions_create(
        self, req: tsi.CompletionsCreateReq
    ) -> tsi.CompletionsCreateRes:
        return self._generic_request(
            "/completions/create",
            req,
            tsi.CompletionsCreateReq,
            tsi.CompletionsCreateRes,
        )


__docspec__ = [
    RemoteHTTPTraceServer,
    ServerInfoRes,
    StartBatchItem,
    EndBatchItem,
    Batch,
]



from typing import Any, Optional

import weave
from weave.integrations.dspy.dspy_utils import (
    dspy_postprocess_inputs,
    dump_dspy_objects,
    get_op_name_for_callback,
)
from weave.trace.context import weave_client_context as weave_client_context
from weave.trace.serialization.serialize import dictify
from weave.trace.weave_client import Call

import_failed = False

try:
    from dspy.utils.callback import BaseCallback
except ImportError:
    import_failed = True

if not import_failed:

    class WeaveCallback(BaseCallback):
        def __init__(self) -> None:
            self._call_map: dict[str, Call] = {}

        def on_module_start(
            self, call_id: str, instance: Any, inputs: dict[str, Any]
        ) -> None:
            gc = weave_client_context.require_weave_client()
            if instance is not None:
                inputs = {"self": dictify(instance), **inputs}
                if hasattr(instance, "signature"):
                    if hasattr(instance.signature, "model_json_schema"):
                        inputs["self"]["signature"] = (
                            instance.signature.model_json_schema()
                        )
                    else:
                        inputs["self"]["signature"] = instance.signature

            op_name = get_op_name_for_callback(instance, inputs)
            self._call_map[call_id] = gc.create_call(
                op_name,
                inputs=dspy_postprocess_inputs(inputs),
                display_name=op_name,
            )

        def on_module_end(
            self,
            call_id: str,
            outputs: Optional[Any],
            exception: Optional[Exception] = None,
        ) -> None:
            gc = weave_client_context.require_weave_client()
            if call_id in self._call_map:
                gc.finish_call(
                    self._call_map[call_id], dump_dspy_objects(outputs), exception
                )

        def on_tool_start(
            self, call_id: str, instance: Any, inputs: dict[str, Any]
        ) -> None:
            gc = weave_client_context.require_weave_client()

            if instance is not None:
                inputs = {"self": dictify(instance), **inputs}

            # TODO: Should we do this? Have to ask Ayush tomorrow.
            if hasattr(instance, "func"):
                instance.func = weave.op(instance.func)

            op_name = get_op_name_for_callback(instance, inputs)
            self._call_map[call_id] = gc.create_call(
                op_name,
                inputs=dspy_postprocess_inputs(inputs),
                display_name=op_name,
            )

        def on_tool_end(
            self,
            call_id: str,
            outputs: Optional[dict[str, Any]],
            exception: Optional[Exception] = None,
        ) -> None:
            gc = weave_client_context.require_weave_client()
            if call_id in self._call_map:
                gc.finish_call(
                    self._call_map[call_id], dump_dspy_objects(outputs), exception
                )



from __future__ import annotations

import logging
from functools import wraps
from typing import Any, Callable, TypeVar

import requests
import tenacity
from pydantic import ValidationError

from weave.trace.settings import retry_max_attempts, retry_max_interval

logger = logging.getLogger(__name__)

T = TypeVar("T")


def with_retry(func: Callable[..., T]) -> Callable[..., T]:
    """
    Decorator that applies configurable retry logic to a function.
    Retry configuration is determined by:
    1. Values from weave.trace.settings (if available)
    2. Values set via configure_retry()
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> T:
        retry = tenacity.Retrying(
            stop=tenacity.stop_after_attempt(retry_max_attempts()),
            wait=tenacity.wait_exponential_jitter(initial=1, max=retry_max_interval()),
            retry=tenacity.retry_if_exception(_is_retryable_exception),
            before_sleep=_log_retry,
            retry_error_callback=_log_failure,
            reraise=True,
        )
        return retry(lambda: func(*args, **kwargs))

    return wrapper


def _is_retryable_exception(e: Exception) -> bool:
    # Don't retry pydantic validation errors
    if isinstance(e, ValidationError):
        return False

    # Don't retry on HTTP 4xx (except 429)
    if isinstance(e, requests.HTTPError) and e.response is not None:
        code_class = e.response.status_code // 100

        # Bad request, not rate-limiting
        if code_class == 4 and e.response.status_code != 429:
            return False

    # Otherwise, retry: 5xx, OSError, ConnectionError, ConnectionResetError, IOError, etc...
    return True


def _log_retry(retry_state: tenacity.RetryCallState) -> None:
    logger.info(
        "retry_attempt",
        extra={
            "fn": retry_state.fn,
            "attempt_number": retry_state.attempt_number,
            "exception": str(retry_state.outcome.exception()),
        },
    )


def _log_failure(retry_state: tenacity.RetryCallState) -> Any:
    logger.info(
        "retry_failed",
        extra={
            "fn": retry_state.fn,
            "attempt_number": retry_state.attempt_number,
            "exception": str(retry_state.outcome.exception()),
        },
    )
    return retry_state.outcome.result()



from typing import Any, Optional

import jsonschema
from pydantic import BaseModel, Field, create_model, field_validator, model_validator
from pydantic.fields import FieldInfo

from weave.trace_server.interface.builtin_object_classes import base_object_def

SUPPORTED_PRIMITIVES = (int, float, bool, str)


class AnnotationSpec(base_object_def.BaseObject):
    field_schema: dict[str, Any] = Field(
        default={},
        description="Expected to be valid JSON Schema. Can be provided as a dict, a Pydantic model class, a tuple of a primitive type and a Pydantic Field, or primitive type",
        examples=[
            # String feedback
            {"type": "string", "maxLength": 100},
            # Number feedback
            {"type": "number", "minimum": 0, "maximum": 100},
            # Integer feedback
            {"type": "integer", "minimum": 0, "maximum": 100},
            # Boolean feedback
            {"type": "boolean"},
            # Categorical feedback
            {"type": "string", "enum": ["option1", "option2"]},
        ],
    )

    # TODO
    # If true, all unique creators will have their
    # own value for this feedback type. Otherwise,
    # by default, the value is shared and can be edited.
    unique_among_creators: bool = False

    # TODO
    # If provided, this feedback type will only be shown
    # when a call is generated from the given op ref
    op_scope: Optional[list[str]] = Field(
        default=None,
        examples=[
            ["weave:///entity/project/op/name:digest"],
            ["weave:///entity/project/op/name:*"],
        ],
    )

    @model_validator(mode="before")
    @classmethod
    def preprocess_field_schema(cls, data: dict[str, Any]) -> dict[str, Any]:
        if "field_schema" not in data:
            return data

        field_schema = data["field_schema"]

        temp_field_tuple = None
        # Handle Pydantic Field
        if isinstance(field_schema, tuple):
            if len(field_schema) != 2:
                raise ValueError("Expected a tuple of length 2")
            annotation, field = field_schema
            if (
                not isinstance(annotation, type)
            ) or annotation not in SUPPORTED_PRIMITIVES:
                raise TypeError("Expected annotation to be a primitive type")
            if not isinstance(field, FieldInfo):
                raise TypeError("Expected field to be a Pydantic Field")
            temp_field_tuple = (annotation, field)
        elif field_schema in SUPPORTED_PRIMITIVES:
            temp_field_tuple = (field_schema, Field())

        if temp_field_tuple is not None:
            # Create a temporary model to leverage Pydantic's schema generation
            TempModel = create_model("TempModel", field=temp_field_tuple)

            schema = TempModel.model_json_schema()["properties"]["field"]

            if (
                "title" in schema and schema["title"] == "Field"
            ):  # default title for Field
                schema.pop("title")

            data["field_schema"] = schema
            return data

        # Handle Pydantic model
        if isinstance(field_schema, type) and issubclass(field_schema, BaseModel):
            data["field_schema"] = field_schema.model_json_schema()  # type: ignore
            return data

        return data

    @field_validator("field_schema")
    def validate_field_schema(cls, schema: dict[str, Any]) -> dict[str, Any]:
        # Validate the schema
        try:
            jsonschema.validate(None, schema)
        except jsonschema.exceptions.SchemaError as e:
            raise e
        except jsonschema.exceptions.ValidationError:
            pass  # we don't care that `None` does not conform
        return schema

    def value_is_valid(self, payload: Any) -> bool:
        """
        Validates a payload against this annotation spec's schema.

        Args:
            payload: The data to validate against the schema

        Returns:
            bool: True if validation succeeds, False otherwise
        """
        try:
            jsonschema.validate(payload, self.field_schema)
        except jsonschema.exceptions.ValidationError:
            return False
        return True



from typing import Any, Optional, TypedDict

from pydantic import BaseModel

from weave.trace_server.interface.builtin_object_classes.builtin_object_registry import (
    BUILTIN_OBJECT_REGISTRY,
)

"""
There are two standard base object classes: BaseObject and Object

`Object` is the base class for the more advanced object-oriented `weave.Object` use cases.
`BaseObject` is the more simple schema-based base object class.
"""
base_object_class_names = ["BaseObject", "Object"]


class GetObjectClassesResult(TypedDict):
    # object_class is the "leaf" class of the val assuming it is a subclass of weave.Object or weave.BaseObject
    object_class: Optional[str]
    # base_object_class is the first subclass of weave.Object or weave.BaseObject
    base_object_class: Optional[str]


def get_object_classes(val: Any) -> Optional[GetObjectClassesResult]:
    if isinstance(val, dict):
        if "_bases" in val:
            if isinstance(val["_bases"], list):
                if len(val["_bases"]) >= 2:
                    if val["_bases"][-1] == "BaseModel":
                        if val["_bases"][-2] in base_object_class_names:
                            object_class = val["_class_name"]
                            base_object_class = object_class
                            if len(val["_bases"]) > 2:
                                base_object_class = val["_bases"][-3]
                            return GetObjectClassesResult(
                                object_class=object_class,
                                base_object_class=base_object_class,
                            )
    return None


class ProcessIncomingObjectResult(TypedDict):
    val: Any
    base_object_class: Optional[str]


def process_incoming_object_val(
    val: Any, req_builtin_object_class: Optional[str] = None
) -> ProcessIncomingObjectResult:
    """
    This method is responsible for accepting an incoming object from the user and validating it
    against the object class. It adds the _class_name and _bases keys correctly and returns the object
    with the base object class set. It does not mutate the original object, but returns a new object
    with values set if needed.
    """
    # First, we ensure the object is a dict before processing it.
    # If the object is not a dict, we return it as is and set the base_object_class to None.
    if not isinstance(val, dict):
        if req_builtin_object_class is not None:
            raise ValueError("object_class cannot be provided for non-dict objects")
        return ProcessIncomingObjectResult(val=val, base_object_class=None)

    # Next we extract the object classes from the object. the `_bases` and `_class_name` keys are
    # special weave-added keys that tell us the class hierarchy of the object.
    # _class_name is the name of the class of the object.
    # _bases is a list of the class's superclasses.
    # In the specific case that bases starts with `BaseModel` (Pydantic Root) and the next subclass
    # is in `base_object_class_names`, then we assume it is a special Weave Object class. From there,
    # we can extract the object class and base object class.
    val_object_classes = get_object_classes(val)

    # In the event that we successfully extracted the object classes, we need to check if the
    # requested object class matches the object class of the object. If it does not, we raise an error.
    if val_object_classes:
        if req_builtin_object_class:
            if val_object_classes["object_class"] != req_builtin_object_class:
                raise ValueError(
                    f"object_class must match val's defined object class: {val_object_classes['object_class']} != {req_builtin_object_class}"
                )
            else:
                # Note: instead of passing here, it is reasonable to conclude that we should instead raise an error -
                # effectively disallowing the user from providing both a requested object class and a class hierarchy inside the payload.
                pass
        # In this case, we assume that the object is valid and do not need to process it.
        # This would happen in practice if the user is editing an existing object by simply modifying the keys.
        return ProcessIncomingObjectResult(
            val=val, base_object_class=val_object_classes["base_object_class"]
        )

    # Next, we check if the user provided an object class. If they did, we need to validate the object
    # and set the correct bases information. This is an important case: the user is asking us to ensure that they payload is valid and
    # stored correctly. We need to validate the payload and write the correct bases information.
    if req_builtin_object_class is not None:
        if builtin_object_class := BUILTIN_OBJECT_REGISTRY.get(
            req_builtin_object_class
        ):
            # TODO: in the next iteration of this code path, this is where we need to actually publish the object
            # using the weave publish API instead of just dumping it.
            dict_val = dump_object(builtin_object_class.model_validate(val))
            new_val_object_classes = get_object_classes(dict_val)
            if not new_val_object_classes:
                raise ValueError(
                    f"Unexpected error: could not get object classes for {dict_val}"
                )
            if new_val_object_classes["object_class"] != req_builtin_object_class:
                raise ValueError(
                    f"Unexpected error: base object class does not match requested object class: {new_val_object_classes['object_class']} != {req_builtin_object_class}"
                )
            return ProcessIncomingObjectResult(
                val=dict_val,
                base_object_class=new_val_object_classes["base_object_class"],
            )
        else:
            raise ValueError(f"Unknown object class: {req_builtin_object_class}")

    # Finally, if there is no requested object class, just return the object as is.
    return ProcessIncomingObjectResult(val=val, base_object_class=None)


# Server-side version of `pydantic_object_record`
def dump_object(val: BaseModel) -> dict:
    cls = val.__class__
    cls_name = val.__class__.__name__
    bases = [c.__name__ for c in cls.mro()[1:-1]]

    dump = {}
    # Order matters here due to the way we calculate the digest!
    # This matches the client
    dump["_type"] = cls_name
    for k in val.model_fields:
        dump[k] = _general_dump(getattr(val, k))
    # yes, this is done twice, to match the client
    dump["_class_name"] = cls_name
    dump["_bases"] = bases
    return dump


def _general_dump(val: Any) -> Any:
    if isinstance(val, BaseModel):
        return dump_object(val)
    elif isinstance(val, dict):
        return {k: _general_dump(v) for k, v in val.items()}
    elif isinstance(val, list):
        return [_general_dump(v) for v in val]
    elif isinstance(val, tuple):
        return tuple(_general_dump(v) for v in val)
    elif isinstance(val, set):
        return {_general_dump(v) for v in val}
    else:
        return val



from typing import Optional

import weave
from weave.flow.util import transpose


def p_r_f1(tp: int, fp: int, fn: int) -> tuple[float, float, float]:
    """
    Compute precision, recall, and F1 score based on true positives (tp), false positives (fp), and false negatives (fn).

    If any denominator is zero, the corresponding metric is set to zero.

    Args:
        tp (int): Number of true positives.
        fp (int): Number of false positives.
        fn (int): Number of false negatives.

    Returns:
        tuple[float, float, float]: A tuple containing (precision, recall, f1) scores.
    """
    precision: float = 0
    if tp or fp:
        precision = tp / (tp + fp)
    recall: float = 0
    if tp or fn:
        recall = tp / (tp + fn)
    f1: float = 0
    if precision or recall:
        f1 = 2 * (precision * recall) / (precision + recall)
    return precision, recall, f1


class MultiTaskBinaryClassificationF1(weave.Scorer):
    """
    Multi-task binary classification scorer that computes precision, recall, and F1 score for each target class based on
    the model output and ground truth labels.

    Attributes:
        class_names (list[str]): The list of target class names.

    Methods:
        score(target: dict, model_output: Optional[dict]) -> dict:
            Compares the target class labels with the model outputs to indicate correctness for each class.
            Uses the "model_output" key for backwards compatibility.

        summarize(score_rows: list) -> Optional[dict]:
            Aggregates multiple scoring results to compute the precision, recall, and F1 score for each class.
    """

    class_names: list[str]

    @weave.op()
    def summarize(self, score_rows: list) -> Optional[dict]:
        """
        Aggregate scoring results and compute precision, recall, and F1 score for each class.

        Args:
            score_rows (list[dict]): A list of score dictionaries for each sample where each dictionary
            contains entries for each class with keys "correct" and "negative".

        Returns:
            Optional[dict]: A dictionary mapping each class name to its metrics,
            including keys "precision", "recall", and "f1".
        """
        result = {}
        cols = transpose(score_rows)

        for class_name in self.class_names:
            col = cols[class_name]
            tp = sum(r["correct"] and not r["negative"] for r in col)
            fp = sum(not r["correct"] and not r["negative"] for r in col)
            fn = sum(not r["correct"] and r["negative"] for r in col)
            precision, recall, f1 = p_r_f1(tp, fp, fn)
            result[class_name] = {"f1": f1, "precision": precision, "recall": recall}

        return result

    # NOTE: This is an old-style scorer that uses `model_output` instead of `output` for
    # backwards compatibility.  In the future, this behavior may change to use the newer `output` key.
    # You can still pass a `column_map` to map to the new `output` key if preferred.
    @weave.op()
    def score(self, target: dict, model_output: Optional[dict]) -> dict:
        """
        Compare target labels with model outputs to determine correctness for each class.

        Args:
            target (dict): A dictionary mapping each class name to the ground truth label.
            model_output (Optional[dict]): A dictionary mapping each class name to the model's output.
                If None, outputs are treated as false.

        Returns:
            dict: A dictionary mapping each class name to a dictionary containing:
                - "correct": True if the target label matches the model output.
                - "negative": True if the model output is missing or false.
        Note:
            This method uses the `model_output` key for backwards compatibility.
        """
        result = {}
        for class_name in self.class_names:
            class_label = target.get(class_name)
            class_output = model_output.get(class_name) if model_output else None
            result[class_name] = {
                "correct": class_label == class_output,
                "negative": not class_output,
            }
        return result



from typing import Any, Union

import numpy as np
from pydantic import Field, validate_call

import weave
from weave.flow.scorer import WeaveScorerResult
from weave.scorers.default_models import MODEL_PATHS
from weave.scorers.scorer_types import HuggingFaceScorer
from weave.scorers.utils import load_hf_model_weights

CONTEXT_RELEVANCE_SCORER_THRESHOLD = 0.55


class WeaveContextRelevanceScorerV1(HuggingFaceScorer):
    """
    A scorer that evaluates the relevance of model outputs relative to input queries and context.
    The scorer uses a fine-tuned deberta-small-long-nli model from tasksource;
    https://huggingface.co/tasksource/deberta-small-long-nli

    This scorer uses a fine-tuned model to analyze whether outputs are semantically relevant to their
    input queries and context. It processes text in chunks and returns both binary relevance flags
    and detailed span-level scores.

    Args:
        model_name_or_path (str): Path or name of model weights to load
        device (str): Device to run model on, defaults to "cpu"
        threshold (float): Threshold for relevance classification, defaults to 0.55
        return_all_spans (bool): Return all spans, defaults to False

    Note: This Scorer's `score` method expects the context to be passed to its `output` parameter as
    a string or list of strings.

    Returns:
        WeaveScorerResult: An object containing:
            - passed (bool): Whether the output was flagged as relevant (score >= threshold)
            - metadata (dict): Contains:
                - score (float): Overall relevance score between 0 and 1
                - all_spans (list[dict], optional): If `return_all_spans` is True, includes list of all relevant
                  text spans with their scores, where each dict has:
                    - text (str): The relevant text span
                    - score (float): The relevance score for this span

    Example:
        >>> scorer = WeaveContextRelevanceScorerV1(return_all_spans=True)
        >>> result = scorer.score(
        ...     query="What is the capital of France?",
        ...     output=["Paris is the capital of France."], # the context to score
        ... )
        >>> print(result)
        WeaveScorerResult(
            passed=True,
            metadata={
                'score': 0.92,
                'all_spans': [
                    {'text': 'Paris is the capital of France', 'score': 0.92}
                ]
            }
        )
    """

    threshold: float = Field(
        default=CONTEXT_RELEVANCE_SCORER_THRESHOLD,
        description="The threshold for relevance classification.",
    )
    return_all_spans: bool = Field(
        default=False,
        description="Whether to return all spans.",
    )
    model_max_length: int = 1280

    def load_model(self) -> None:
        from transformers import AutoModelForTokenClassification

        self._local_model_path = load_hf_model_weights(
            self.model_name_or_path, MODEL_PATHS["relevance_scorer"]
        )
        self._model = AutoModelForTokenClassification.from_pretrained(
            self._local_model_path
        ).to(self.device)
        self._model.eval()

    def load_tokenizer(self) -> None:
        from transformers import AutoTokenizer

        self._tokenizer = AutoTokenizer.from_pretrained(
            self._local_model_path,
            model_max_length=self.model_max_length,
        )

    def _score_document(
        self, query: str, document: str, threshold: float
    ) -> tuple[list[dict[str, Any]], int, int]:
        """Score a single document."""
        import torch

        assert self._tokenizer is not None  # keep mypy happy
        assert self._model is not None  # keep mypy happy

        input_text = query + f" {self._tokenizer.sep_token} " + document
        model_inputs = self._tokenizer(
            input_text,
            truncation=True,
            max_length=self.model_max_length,
            padding=False,
            return_tensors="pt",
            return_special_tokens_mask=True,
        )

        model_inputs = {k: v.to(self.device) for k, v in model_inputs.items()}

        special_tokens_mask = model_inputs.pop("special_tokens_mask")
        combined_mask = (
            ~((model_inputs["input_ids"] == 2).bool() | special_tokens_mask.bool())
            .cpu()
            .numpy()
            .flatten()
        )
        # we should mask the query up to the sep token,
        # on the combined mask we have to search for the first False
        # TODO: Check that this is not wrong
        false_indices = np.where(~combined_mask)[0]
        start = false_indices[0]
        end = false_indices[1]
        combined_mask[start:end] = False

        with torch.inference_mode():
            results = self._model(**model_inputs)
            logits = results.logits[0].detach()
            probabilities = torch.nn.functional.softmax(logits, dim=-1).detach()

        pred_mask = (
            (probabilities[:, 1] > threshold).cpu().numpy().astype(int).flatten()
        )
        label_mask = pred_mask & combined_mask

        positive_probs = probabilities[:, 1].cpu().numpy()
        transitions = np.diff(np.concatenate([[0], label_mask, [0]]))
        starts = np.where(transitions == 1)[0]
        ends = np.where(transitions == -1)[0]

        spans_with_probs = []
        token_ids = model_inputs["input_ids"].cpu().numpy()[0]

        for start, end in zip(starts, ends):
            span_text = self._tokenizer.decode(token_ids[start:end])
            span_prob = positive_probs[start:end].mean()
            spans_with_probs.append({"text": span_text, "score": float(span_prob)})

        return spans_with_probs, int(label_mask.sum()), int(len(label_mask))

    @validate_call
    @weave.op
    def score(
        self,
        query: str,
        output: Union[str, list[str]],  # Pass the context to the `output` parameter
    ) -> WeaveScorerResult:
        all_spans: list[dict[str, Any]] = []
        total_weighted_score = 0.0
        total_length = 0

        if isinstance(output, str):
            output = [output]
        for doc in output:
            spans, relevant_tokens, total_tokens = self._score_document(
                query, doc, self.threshold
            )

            all_spans.extend(spans)

            if total_tokens > 0:
                doc_score = relevant_tokens / total_tokens
                doc_weight = total_tokens
                total_weighted_score += doc_score * doc_weight
                total_length += total_tokens

        final_score = total_weighted_score / total_length if total_length > 0 else 0.0
        metadata: dict[str, Any] = {"score": final_score}
        if self.return_all_spans:
            metadata["all_spans"] = all_spans
        return WeaveScorerResult(
            passed=final_score >= self.threshold,
            metadata=metadata,
        )
